"""ORBIT specific marmot.Environment."""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


from bisect import bisect

import numpy as np
from marmot import Environment
from marmot._core import Constraint
from numpy.lib.recfunctions import append_fields


class OrbitEnvironment(Environment):
    """ORBIT Specific Environment."""

    def __init__(self, name="Environment", state=None, **kwargs):
        """
        Creates an instance of Environment.

        Parameters
        ----------
        name : str
            Environment name.
            Default: 'Environment'
        state : array-like
            Time series representing the state of the environment throughout
            time or iterations.
        """

        super().__init__()

        self.name = name
        self.state = self.standarize_state_inputs(state)
        self.alpha = kwargs.get("ws_alpha", 0.1)
        self.default_height = kwargs.get("ws_default_height", 10)

        self._logs = []
        self._agents = {}
        self._objects = []

    def _find_valid_constraints(self, **kwargs):
        """
        Finds any constraitns in `kwargs` where the key matches a column name
        in `self.state` and the value type is `Constraint`.

        This method overrides the default method to handle windspeed
        constraints specifically, interpolating or extrapolating environment
        windspeed profiles to the input constraint height.

        Returns
        -------
        valid : dict
            Valid constraints that apply to a column in `self.state`.
        """

        c = {k: v for k, v in kwargs.items() if isinstance(v, Constraint)}
        constraints = self.resolve_windspeed_constraints(c)

        keys = set(self.state.dtype.names).intersection(set(constraints.keys()))
        valid = {k: v for k, v in constraints.items() if k in keys}

        return valid

    def standarize_state_inputs(self, _in):
        """
        Standardization routines applied to columns in `self.state`.

        Parameters
        ----------
        _in : array-like
            Time series representing the state of the environment throughout
            time or iterations.

        Returns
        -------
        state : array-like
            Time series with windspeed heights in simplest representation.
        """

        if _in is None:
            return None

        names = []
        for name in list(_in.dtype.names):

            if "windspeed" in name:
                try:
                    val = name.split("_")[1].replace("m", "")
                    new = f"windspeed_{self.simplify_num(val)}m"
                    names.append(new)

                except IndexError:
                    names.append(name)

            else:
                names.append(name)

        state = _in.copy()
        state.dtype.names = names
        return state

    def resolve_windspeed_constraints(self, constraints):
        """
        Resolves the applied windspeed constraints given the windspeed profiles
        that are present in `self.state`.

        Parameters
        ----------
        constraints : dict
            Dictionary of constraints

        Returns
        -------
        constraitns : dict
            Dictionary of constraints with windspeed columns resolved to their
            simplest representation.
        """

        ws = {}
        for k in list(constraints.keys()):
            if "windspeed" in k:
                ws[k] = constraints.pop(k)

        if not ws:
            return constraints

        if "windspeed" in self.state.dtype.names:
            if len(ws) > 1:
                raise ValueError("Multiple constraints applied to the 'windspeed' column.")

            return {**constraints, "windspeed": list(ws.values())[0]}

        for k, v in ws.items():

            if k == "windspeed":
                height = self.simplify_num(self.default_height)

            else:
                val = k.split("_")[1].replace("m", "")
                height = self.simplify_num(val)

            name = f"windspeed_{height}m"
            if name in self.state.dtype.names:
                pass

            else:
                loc = bisect(self.ws_heights, height)
                if loc == 0:
                    self.extrapolate_ws(self.ws_heights[0], height)

                elif loc == len(self.ws_heights):
                    self.extrapolate_ws(self.ws_heights[-1], height)

                else:
                    h1 = self.ws_heights[loc - 1]
                    h2 = self.ws_heights[loc]
                    self.interpolate_ws(h1, h2, height)

            constraints[name] = v

        return constraints

    @property
    def ws_heights(self):
        """Returns heights of available windspeed profiles."""

        columns = [c for c in self.state.dtype.names if "windspeed" in c]

        heights = []
        for c in columns:
            try:
                val = c.split("_")[1].replace("m", "")
                heights.append(float(val) if "." in val else int(val))

            except IndexError:
                pass

        return sorted(heights)

    def interpolate_ws(self, h1, h2, h):
        """
        Interpolates between two windspeed profiles using the power law. This
        method will calculate the power law coefficient between `h1` and `h2`.

        Parameters
        ----------
        h1 : int | float
            Lower measurement height.
        h2 : int | float
            Upper measurement height.
        h : int | float
            Desired profile height.
        """

        ts1 = self.state[f"windspeed_{h1}m"]
        ts2 = self.state[f"windspeed_{h2}m"]
        alpha = np.log(ts2.mean() / ts1.mean()) / np.log(h2 / h1)

        ts = ts1 * (h / h1) ** alpha

        self.state = np.array(append_fields(self.state, f"windspeed_{h}m", ts))

    def extrapolate_ws(self, h1, h):
        """
        Extrapolates a windspeed profile using the power law coefficient
        at `self.alpha`.

        Parameters
        ----------
        h1 : int | float
            Measurement height.
        h : int | float
            Desired profile height.
        """

        ts1 = self.state[f"windspeed_{h1}m"]
        ts = ts1 * (h / h1) ** self.alpha

        self.state = np.array(append_fields(self.state, f"windspeed_{h}m", ts))

    @staticmethod
    def simplify_num(str):
        """Returns the simplest str representation of a number."""

        num = float(str)
        if int(num) == num:
            return int(num)

        else:
            return num



"""`FloatingSubstationInstallation` and related processes."""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2021, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


from marmot import Agent, le, process
from marmot._exceptions import AgentNotRegistered

from wisdem.orbit.core import WetStorage
from wisdem.orbit.core.logic import position_onsite
from wisdem.orbit.phases.install import InstallPhase
from wisdem.orbit.phases.install.mooring_install.mooring import (
    install_mooring_line,
    install_mooring_anchor,
    perform_mooring_site_survey,
)


class FloatingSubstationInstallation(InstallPhase):
    """
    Offshore Substation (OSS) installation process using the quayside assembly
    and tow-out processes.
    """

    phase = "Offshore Substation Installation"
    capex_category = "Offshore Substation"

    #:
    expected_config = {
        "num_substations": "int",
        "oss_install_vessel": "str",
        "site": {"distance": "km", "depth": "m"},
        "offshore_substation_topside": {
            "unit_cost": "USD",
            "attach_time": "int | float (optional, default: 24)",
        },
        "offshore_substation_substructure": {
            "type": "Floating",
            "takt_time": "int | float (optional, default: 0)",
            "unit_cost": "USD",
            "mooring_cost": "USD",
            "towing_speed": "int | float (optional, default: 6 km/h)",
        },
    }

    def __init__(self, config, weather=None, **kwargs):
        """
        Creates an instance of OffshoreSubstationInstallation.

        Parameters
        ----------
        config : dict
            Simulation specific configuration.
        weather : np.ndarray
            Weather profile at site.
        """

        super().__init__(weather, **kwargs)

        config = self.initialize_library(config, **kwargs)
        self.config = self.validate_config(config)

        self.initialize_port()
        self.setup_simulation(**kwargs)

    def setup_simulation(self, **kwargs):
        """
        Initializes required objects for simulation.
        - Creates port

        - Creates support vessel + towing vessels
        """

        self.distance = self.config["site"]["distance"]
        self.num_substations = self.config["num_substations"]

        self.initialize_substructure_production()
        self.initialize_installation_vessel()

    @property
    def system_capex(self):
        """Returns total procurement cost of the substation substructures,
        topsides and mooring."""

        topside = self.config["offshore_substation_topside"]["unit_cost"]
        substructure = self.config["offshore_substation_substructure"]["unit_cost"]
        mooring = self.config["offshore_substation_substructure"]["mooring_cost"]

        return self.num_substations * (topside + substructure + mooring)

    def initialize_substructure_production(self):
        """
        Initializes the production of the floating substation substructures at
        quayside.
        """

        self.wet_storage = WetStorage(self.env, float("inf"))
        takt_time = self.config["offshore_substation_substructure"].get("takt_time", 0)
        attach_time = self.config["offshore_substation_topside"].get("attach_time", 24)
        to_assemble = [1] * self.num_substations

        self.assembly_line = SubstationAssemblyLine(to_assemble, takt_time, attach_time, self.wet_storage, 1)

        self.env.register(self.assembly_line)
        self.assembly_line.start()

    def initialize_installation_vessel(self):
        """Initialize the floating substation installation vessel."""

        support = self.config["oss_install_vessel"]
        vessel = self.initialize_vessel("Floating Substation Installation Vessel", support)
        self.env.register(vessel)
        vessel.initialize(mobilize=False)
        self.support_vessel = vessel

        depth = self.config["site"]["depth"]
        towing_speed = self.config["offshore_substation_substructure"].get("towing_speed", 6)

        install_floating_substations(
            self.support_vessel,
            self.wet_storage,
            self.distance,
            towing_speed,
            depth,
            self.num_substations,
        )

    @property
    def detailed_output(self):

        return {}


@process
def install_floating_substations(vessel, feed, distance, towing_speed, depth, number):
    """
    Process steps that installation vessel at site performs to install floating
    substations.

    Parameters
    ----------
    vessel : Agent
        Performing agent.
    feed : simply.Resource
        Wet storage for completed assemblies.
    distance : int | float
        Distance from port to site.
    towing_speed : int | float
        Speed at which completed assembly can be towed to site at (km/h).
    depth : int | float
        Site depth (m).
    number : int
        Number of substations to install.
    """

    travel_time = distance / towing_speed

    for _ in range(number):

        start = vessel.env.now
        yield feed.get()
        delay = vessel.env.now - start
        if delay > 0:
            vessel.submit_action_log("Delay: Waiting on Completed Assembly", delay)

        yield vessel.task(
            "Tow Substation to Site",
            travel_time,
            constraints=vessel.operational_limits,
        )
        yield position_onsite(vessel)
        yield vessel.task_wrapper(
            "Ballast to Operational Draft",
            6,
            constraints={"windspeed": le(15), "waveheight": le(2.5)},
        )

        for _ in range(3):
            yield perform_mooring_site_survey(vessel)
            yield install_mooring_anchor(vessel, depth, "Suction Pile")
            yield install_mooring_line(vessel, depth)
            yield vessel.task_wrapper(
                "Connect Mooring Lines",
                22,
                suspendable=True,
                constraints={"windspeed": le(15), "waveheight": le(2.5)},
            )
            yield vessel.task_wrapper(
                "Check Mooring Lines",
                12,
                suspendable=True,
                constraints={"windspeed": le(15), "waveheight": le(2.5)},
            )

        yield vessel.transit(distance)


class SubstationAssemblyLine(Agent):
    """Substation Assembly Line Class."""

    def __init__(self, assigned, takt_time, attach_time, target, num):
        """
        Creates an instance of `SubstructureAssemblyLine`.

        Parameters
        ----------
        assigned : list
            List of assigned tasks. Can be shared with other assembly lines.
        takt_time : int | float
            Hours required to produce one substructure.
        attach_time : int | float
            Hours required to attach a topside to the substructure.
        target : simpy.Store
            Target storage.
        num : int
            Assembly line number designation.
        """

        super().__init__(f"Substation Assembly Line {num}")

        self.assigned = assigned
        self.takt_time = takt_time
        self.attach_time = attach_time
        self.target = target

    def submit_action_log(self, action, duration, **kwargs):
        """
        Submits a log representing a completed `action` performed over time
        `duration`.

        This method overwrites the default `submit_action_log` in
        `marmot.Agent`, adding operation cost to every submitted log within
        ORBIT.

        Parameters
        ----------
        action : str
            Performed action.
        duration : int | float
            Duration of action.

        Raises
        ------
        AgentNotRegistered
        """

        if self.env is None:
            raise AgentNotRegistered(self)

        else:
            payload = {
                **kwargs,
                "agent": str(self),
                "action": action,
                "duration": float(duration),
                "cost": 0,
            }

            self.env._submit_log(payload, level="ACTION")

    @process
    def assemble_substructure(self):
        """
        Simulation process for assembling a substructure.
        """

        yield self.task("Substation Substructure Assembly", self.takt_time)
        yield self.task("Attach Topside", self.attach_time)
        substation = FloatingSubstation()

        start = self.env.now
        yield self.target.put(substation)
        delay = self.env.now - start

        if delay > 0:
            self.submit_action_log("Delay: No Wet Storage Available", delay)

    @process
    def start(self):
        """
        Trigger the assembly line to run. Will attempt to pull a task from
        self.assigned and timeout for the assembly time. Shuts down after
        self.assigned is empty.
        """

        while True:
            try:
                _ = self.assigned.pop(0)
                yield self.assemble_substructure()

            except IndexError:
                break


class FloatingSubstation:
    """Floating Substructure Class."""

    def __init__(self):
        """Creates an instance of `Substructure`."""

        pass



import openmdao.api as om

from wisdem.floatingse.mooring import Mooring
from wisdem.floatingse.constraints import RigidModes, FloatingConstraints
from wisdem.commonse.cylinder_member import MemberDetailed
from wisdem.floatingse.floating_frame import FloatingFrame
from wisdem.floatingse.floating_system import FloatingSystem


class FloatingSE(om.Group):
    def initialize(self):
        self.options.declare("modeling_options")

    def setup(self):
        opt = self.options["modeling_options"]

        n_member = opt["floating"]["members"]["n_members"]
        mem_prom = [
            "E_mat",
            "G_mat",
            "sigma_y_mat",
            "sigma_ult_mat",
            "wohler_exp_mat",
            "wohler_A_mat",
            "rho_mat",
            "rho_water",
            "unit_cost_mat",
            "material_names",
            "painting_cost_rate",
            "labor_cost_rate",
        ]
        for k in range(n_member):
            self.add_subsystem(
                f"member{k}",
                MemberDetailed(
                    column_options=opt["floating"]["members"],
                    idx=k,
                    n_mat=opt["materials"]["n_mat"],
                    memmax=True,
                    n_refine=2,
                ),
                promotes=mem_prom + [("joint1", f"member{k}:joint1"), ("joint2", f"member{k}:joint2")],
            )

        # Combine all members and tower into single system
        self.add_subsystem("sys", FloatingSystem(modeling_options=opt), promotes=["*"])

        self.add_subsystem(
            "mm", Mooring(options=opt["mooring"], gamma=opt["WISDEM"]["FloatingSE"]["gamma_f"]), promotes=["*"]
        )

        # Do the load analysis over one or more load cases
        self.add_subsystem("load", FloatingFrame(modeling_options=opt), promotes=["*"])

        # Evaluate system constraints
        self.add_subsystem("cons", FloatingConstraints(modeling_options=opt), promotes=["*"])

        # Evaluate system constraints
        self.add_subsystem("modal", RigidModes(), promotes=["*"])

        # Connect all input variables from all models
        mem_vars = [
            "section_D",
            "section_t",
            "section_A",
            "section_Asx",
            "section_Asy",
            "section_Ixx",
            "section_Iyy",
            "section_J0",
            "section_rho",
            "section_E",
            "section_G",
            "section_sigma_y",
            "idx_cb",
            "variable_ballast_capacity",
            "variable_ballast_Vpts",
            "variable_ballast_spts",
            "constr_ballast_capacity",
            "buoyancy_force",
            "displacement",
            "center_of_buoyancy",
            "center_of_mass",
            "ballast_mass",
            "total_mass",
            "total_cost",
            "I_total",
            "Awater",
            "Iwater",
            "added_mass",
            "waterline_centroid",
        ]
        mem_load_vars = ["z_global", "d_full", "s_full", "s_all"]

        for k in range(n_member):
            for var in mem_vars:
                self.connect(f"member{k}.{var}", f"member{k}:{var}")

            self.connect(f"member{k}.nodes_xyz_all", f"member{k}:nodes_xyz")
            self.connect(f"member{k}.nodes_r_all", f"member{k}:nodes_r")

            for var in mem_load_vars:
                self.connect(f"member{k}.{var}", f"memload{k}.{var}")



import math

import numpy as np
import openmdao.api as om
from wisdem.commonse.akima import Akima
from wisdem.commonse.csystem import DirectionVector
from wisdem.commonse.utilities import cosd, sind  # , linspace_with_deriv, interp_with_deriv, hstack, vstack
from wisdem.commonse.environment import LogWind, PowerWind, LinearWaves

# -----------------
#  Helper Functions
# -----------------
# "Experiments on the Flow Past a Circular Cylinder at Very High Reynolds Numbers", Roshko
Re_pt = [
    0.00001,
    0.0001,
    0.0010,
    0.0100,
    0.0200,
    0.1220,
    0.2000,
    0.3000,
    0.4000,
    0.5000,
    1.0000,
    1.5000,
    2.0000,
    2.5000,
    3.0000,
    3.5000,
    4.0000,
    5.0000,
    10.0000,
]
cd_pt = [
    4.0000,
    2.0000,
    1.1100,
    1.1100,
    1.2000,
    1.2000,
    1.1700,
    0.9000,
    0.5400,
    0.3100,
    0.3800,
    0.4600,
    0.5300,
    0.5700,
    0.6100,
    0.6400,
    0.6700,
    0.7000,
    0.7000,
]

drag_spline = Akima(np.log10(Re_pt), cd_pt, delta_x=0.0)  # exact akima because control points do not change


def cylinderDrag(Re):
    """Drag coefficient for a smooth circular cylinder.

    Parameters
    ----------
    Re : array_like
        Reynolds number

    Returns
    -------
    cd : array_like
        drag coefficient (normalized by cylinder diameter)

    """

    ReN = Re / 1.0e6

    cd = np.zeros_like(Re)
    dcd_dRe = np.zeros_like(Re)
    idx = ReN > 0
    cd[idx], dcd_dRe[idx], _, _ = drag_spline.interp(np.log10(ReN[idx]))
    dcd_dRe[idx] /= Re[idx] * math.log(10)  # chain rule

    return cd, dcd_dRe


# -----------------
#  Components
# -----------------


class AeroHydroLoads(om.ExplicitComponent):
    """
    Compute summed forces due to wind and wave loads.

    Parameters
    ----------
    windLoads_Px : numpy array[nPoints], [N/m]
        distributed loads, force per unit length in x-direction
    windLoads_Py : numpy array[nPoints], [N/m]
        distributed loads, force per unit length in y-direction
    windLoads_Pz : numpy array[nPoints], [N/m]
        distributed loads, force per unit length in z-direction
    windLoads_qdyn : numpy array[nPoints], [N/m**2]
        dynamic pressure
    windLoads_z : numpy array[nPoints], [m]
        corresponding heights
    windLoads_beta : float, [deg]
        wind/wave angle relative to inertia c.s.
    waveLoads_Px : numpy array[nPoints], [N/m]
        distributed loads, force per unit length in x-direction
    waveLoads_Py : numpy array[nPoints], [N/m]
        distributed loads, force per unit length in y-direction
    waveLoads_Pz : numpy array[nPoints], [N/m]
        distributed loads, force per unit length in z-direction
    waveLoads_qdyn : numpy array[nPoints], [N/m**2]
        dynamic pressure
    waveLoads_z : numpy array[nPoints], [m]
        corresponding heights
    waveLoads_beta : float, [deg]
        wind/wave angle relative to inertia c.s.
    z : numpy array[nPoints], [m]
        locations along cylinder
    yaw : float, [deg]
        yaw angle

    Returns
    -------
    Px : numpy array[nPoints], [N/m]
        force per unit length in x-direction
    Py : numpy array[nPoints], [N/m]
        force per unit length in y-direction
    Pz : numpy array[nPoints], [N/m]
        force per unit length in z-direction
    qdyn : numpy array[nPoints], [N/m**2]
        dynamic pressure

    """

    def initialize(self):
        self.options.declare("nPoints")

    def setup(self):
        nPoints = self.options["nPoints"]

        self.add_input("windLoads_Px", np.zeros(nPoints), units="N/m")
        self.add_input("windLoads_Py", np.zeros(nPoints), units="N/m")
        self.add_input("windLoads_Pz", np.zeros(nPoints), units="N/m")
        self.add_input("windLoads_qdyn", np.zeros(nPoints), units="N/m**2")
        self.add_input("windLoads_z", np.zeros(nPoints), units="m")
        self.add_input("windLoads_beta", 0.0, units="deg")
        self.add_input("waveLoads_Px", np.zeros(nPoints), units="N/m")
        self.add_input("waveLoads_Py", np.zeros(nPoints), units="N/m")
        self.add_input("waveLoads_Pz", np.zeros(nPoints), units="N/m")
        self.add_input("waveLoads_qdyn", np.zeros(nPoints), units="N/m**2")
        self.add_input("waveLoads_z", np.zeros(nPoints), units="m")
        self.add_input("waveLoads_beta", 0.0, units="deg")
        self.add_input("z", np.zeros(nPoints), units="m")
        self.add_input("yaw", 0.0, units="deg")

        self.add_output("Px", np.zeros(nPoints), units="N/m")
        self.add_output("Py", np.zeros(nPoints), units="N/m")
        self.add_output("Pz", np.zeros(nPoints), units="N/m")
        self.add_output("qdyn", np.zeros(nPoints), units="N/m**2")

    def compute(self, inputs, outputs):
        z = inputs["z"]
        windLoads = (
            DirectionVector(inputs["windLoads_Px"], inputs["windLoads_Py"], inputs["windLoads_Pz"])
            .inertialToWind(inputs["windLoads_beta"])
            .windToYaw(inputs["yaw"])
        )
        waveLoads = (
            DirectionVector(inputs["waveLoads_Px"], inputs["waveLoads_Py"], inputs["waveLoads_Pz"])
            .inertialToWind(inputs["waveLoads_beta"])
            .windToYaw(inputs["yaw"])
        )

        Px = np.interp(z, inputs["windLoads_z"], windLoads.x) + np.interp(z, inputs["waveLoads_z"], waveLoads.x)
        Py = np.interp(z, inputs["windLoads_z"], windLoads.y) + np.interp(z, inputs["waveLoads_z"], waveLoads.y)
        Pz = np.interp(z, inputs["windLoads_z"], windLoads.z) + np.interp(z, inputs["waveLoads_z"], waveLoads.z)
        qdyn = np.interp(z, inputs["windLoads_z"], inputs["windLoads_qdyn"]) + np.interp(
            z, inputs["waveLoads_z"], inputs["waveLoads_qdyn"]
        )

        # The following are redundant, at one point we will consolidate them to something that works for both cylinder (not using vartrees) and jacket (still using vartrees)
        outputs["Px"] = Px
        outputs["Py"] = Py
        outputs["Pz"] = Pz
        outputs["qdyn"] = qdyn


# -----------------


class CylinderWindDrag(om.ExplicitComponent):
    """
    Compute drag forces on a cylindrical cylinder due to wind.

    Parameters
    ----------
    U : numpy array[nPoints], [m/s]
        magnitude of wind speed
    z : numpy array[nPoints], [m]
        heights where wind speed was computed
    d : numpy array[nPoints], [m]
        corresponding diameter of cylinder section
    beta_wind : float, [deg]
        corresponding wind angles relative to inertial coordinate system
    rho_air : float, [kg/m**3]
        air density
    mu_air : float, [kg/(m*]
        dynamic viscosity of air
    cd_usr : float
        User input drag coefficient to override Reynolds number based one

    Returns
    -------
    windLoads_Px : numpy array[nPoints], [N/m]
        distributed loads, force per unit length in x-direction
    windLoads_Py : numpy array[nPoints], [N/m]
        distributed loads, force per unit length in y-direction
    windLoads_Pz : numpy array[nPoints], [N/m]
        distributed loads, force per unit length in z-direction
    windLoads_qdyn : numpy array[nPoints], [N/m**2]
        dynamic pressure
    windLoads_z : numpy array[nPoints], [m]
        corresponding heights
    windLoads_beta : float, [deg]
        wind/wave angle relative to inertia c.s.

    """

    def initialize(self):
        self.options.declare("nPoints")

    def setup(self):
        nPoints = self.options["nPoints"]

        # variables
        self.add_input("U", np.zeros(nPoints), units="m/s")
        self.add_input("z", np.zeros(nPoints), units="m")
        self.add_input("d", np.zeros(nPoints), units="m")
        self.add_input("beta_wind", 0.0, units="deg")
        self.add_input("rho_air", 0.0, units="kg/m**3")
        self.add_input("mu_air", 0.0, units="kg/(m*s)")
        self.add_input("cd_usr", -1.0)

        self.add_output("windLoads_Px", np.zeros(nPoints), units="N/m")
        self.add_output("windLoads_Py", np.zeros(nPoints), units="N/m")
        self.add_output("windLoads_Pz", np.zeros(nPoints), units="N/m")
        self.add_output("windLoads_qdyn", np.zeros(nPoints), units="N/m**2")
        self.add_output("windLoads_z", np.zeros(nPoints), units="m")
        self.add_output("windLoads_beta", 0.0, units="deg")

        arange = np.arange(nPoints)
        self.declare_partials("windLoads_Px", "U", rows=arange, cols=arange)
        self.declare_partials("windLoads_Px", "d", rows=arange, cols=arange)

        self.declare_partials("windLoads_Py", "U", rows=arange, cols=arange)
        self.declare_partials("windLoads_Py", "d", rows=arange, cols=arange)
        self.declare_partials(["windLoads_Px", "windLoads_Py"], "cd_usr", method="fd")

        self.declare_partials("windLoads_qdyn", "U", rows=arange, cols=arange)
        self.declare_partials("windLoads_qdyn", "rho_air", method="fd")

        self.declare_partials("windLoads_z", "z", rows=arange, cols=arange, val=1.0)
        self.declare_partials("windLoads_beta", "beta_wind", val=1.0)

    def compute(self, inputs, outputs):

        rho = inputs["rho_air"]
        U = inputs["U"]
        d = inputs["d"]
        mu = inputs["mu_air"]
        beta = inputs["beta_wind"]

        # dynamic pressure
        q = 0.5 * rho * U ** 2

        # Reynolds number and drag
        if float(inputs["cd_usr"]) < 0.0:
            Re = rho * U * d / mu
            cd, dcd_dRe = cylinderDrag(Re)
        else:
            cd = inputs["cd_usr"]
            Re = 1.0
            dcd_dRe = 0.0
        Fp = q * cd * d

        # components of distributed loads
        Px = Fp * cosd(beta)
        Py = Fp * sind(beta)
        Pz = 0 * Fp

        # pack data
        outputs["windLoads_Px"] = Px
        outputs["windLoads_Py"] = Py
        outputs["windLoads_Pz"] = Pz
        outputs["windLoads_qdyn"] = q
        outputs["windLoads_z"] = inputs["z"]
        outputs["windLoads_beta"] = beta

    def compute_partials(self, inputs, J):

        # rename
        rho = inputs["rho_air"]
        U = inputs["U"]
        d = inputs["d"]
        mu = inputs["mu_air"]
        beta = inputs["beta_wind"]

        # dynamic pressure
        q = 0.5 * rho * U ** 2

        # Reynolds number and drag
        if float(inputs["cd_usr"]) < 0.0:
            Re = rho * U * d / mu
            cd, dcd_dRe = cylinderDrag(Re)
        else:
            cd = inputs["cd_usr"]
            Re = 1.0
            dcd_dRe = 0.0

        # derivatives
        dq_dU = rho * U
        const = (dq_dU * cd + q * dcd_dRe * rho * d / mu) * d
        dPx_dU = const * cosd(beta)
        dPy_dU = const * sind(beta)

        const = (cd + dcd_dRe * Re) * q
        dPx_dd = const * cosd(beta)
        dPy_dd = const * sind(beta)

        J["windLoads_Px", "U"] = dPx_dU
        J["windLoads_Px", "d"] = dPx_dd

        J["windLoads_Py", "U"] = dPy_dU
        J["windLoads_Py", "d"] = dPy_dd

        J["windLoads_qdyn", "U"] = dq_dU


# -----------------


class CylinderWaveDrag(om.ExplicitComponent):
    """
    Compute drag forces on a cylindrical cylinder due to waves.

    Parameters
    ----------
    U : numpy array[nPoints], [m/s]
        magnitude of wave speed
    A : numpy array[nPoints], [m/s**2]
        magnitude of wave acceleration
    p : numpy array[nPoints], [N/m**2]
        pressure oscillation
    z : numpy array[nPoints], [m]
        heights where wave speed was computed
    d : numpy array[nPoints], [m]
        corresponding diameter of cylinder section
    beta_wave : float, [deg]
        corresponding wave angles relative to inertial coordinate system
    rho_water : float, [kg/m**3]
        water density
    mu_water : float, [kg/(m*]
        dynamic viscosity of water
    cm : float
        mass coefficient
    cd_usr : float
        User input drag coefficient to override Reynolds number based one

    Returns
    -------
    waveLoads_Px : numpy array[nPoints], [N/m]
        distributed loads, force per unit length in x-direction
    waveLoads_Py : numpy array[nPoints], [N/m]
        distributed loads, force per unit length in y-direction
    waveLoads_Pz : numpy array[nPoints], [N/m]
        distributed loads, force per unit length in z-direction
    waveLoads_qdyn : numpy array[nPoints], [N/m**2]
        dynamic pressure
    waveLoads_pt : numpy array[nPoints], [N/m**2]
        total (static+dynamic) pressure
    waveLoads_z : numpy array[nPoints], [m]
        corresponding heights
    waveLoads_beta : float, [deg]
        wind/wave angle relative to inertia c.s.

    """

    def initialize(self):
        self.options.declare("nPoints")

    def setup(self):
        nPoints = self.options["nPoints"]

        # variables
        self.add_input("U", np.zeros(nPoints), units="m/s")
        self.add_input("A", np.zeros(nPoints), units="m/s**2")
        self.add_input("p", np.zeros(nPoints), units="N/m**2")
        self.add_input("z", np.zeros(nPoints), units="m")
        self.add_input("d", np.zeros(nPoints), units="m")
        self.add_input("beta_wave", 0.0, units="deg")
        self.add_input("rho_water", 0.0, units="kg/m**3")
        self.add_input("mu_water", 0.0, units="kg/(m*s)")
        self.add_input("cm", 0.0)
        self.add_input("cd_usr", -1.0)

        self.add_output("waveLoads_Px", np.zeros(nPoints), units="N/m")
        self.add_output("waveLoads_Py", np.zeros(nPoints), units="N/m")
        self.add_output("waveLoads_Pz", np.zeros(nPoints), units="N/m")
        self.add_output("waveLoads_qdyn", np.zeros(nPoints), units="N/m**2")
        self.add_output("waveLoads_pt", np.zeros(nPoints), units="N/m**2")
        self.add_output("waveLoads_z", np.zeros(nPoints), units="m")
        self.add_output("waveLoads_beta", 0.0, units="deg")

        self.declare_partials("*", "rho_water", method="fd")

        arange = np.arange(nPoints)
        self.declare_partials(["waveLoads_Px", "waveLoads_Py"], ["U", "d", "cm", "cd_usr", "beta_wave"], method="fd")
        self.declare_partials("waveLoads_Px", "A", rows=arange, cols=arange)

        self.declare_partials("waveLoads_Py", "A", rows=arange, cols=arange)

        self.declare_partials("waveLoads_qdyn", "U", rows=arange, cols=arange)
        self.declare_partials("waveLoads_pt", "U", rows=arange, cols=arange)
        self.declare_partials("waveLoads_pt", "p", rows=arange, cols=arange, val=1.0)
        self.declare_partials("waveLoads_z", "z", rows=arange, cols=arange, val=1.0)
        self.declare_partials("waveLoads_beta", "beta_wave", val=1.0)

    def compute(self, inputs, outputs):

        # wlevel = inputs['wlevel']
        # if wlevel > 0.0: wlevel *= -1.0

        rho = inputs["rho_water"]
        U = inputs["U"]
        # U0 = inputs['U0']
        d = inputs["d"]
        # zrel= inputs['z']-wlevel
        mu = inputs["mu_water"]
        beta = inputs["beta_wave"]
        # beta0 = inputs['beta0']

        # dynamic pressure
        q = 0.5 * rho * U * np.abs(U)
        # q0= 0.5*rho*U0**2

        # Reynolds number and drag
        if float(inputs["cd_usr"]) < 0.0:
            Re = rho * U * d / mu
            cd, dcd_dRe = cylinderDrag(Re)
        else:
            cd = inputs["cd_usr"] * np.ones_like(d)
            Re = 1.0
            dcd_dRe = 0.0

        # inertial and drag forces
        Fi = rho * inputs["cm"] * math.pi / 4.0 * d ** 2 * inputs["A"]  # Morrison's equation
        Fd = q * cd * d
        Fp = Fi + Fd

        # components of distributed loads
        Px = Fp * cosd(beta)
        Py = Fp * sind(beta)
        Pz = 0.0 * Fp

        # FORCES [N/m] AT z=0 m
        # idx0 = np.abs(zrel).argmin()  # closest index to z=0, used to find d at z=0
        # d0 = d[idx0]  # initialize
        # cd0 = cd[idx0]  # initialize
        # if (zrel[idx0]<0.) and (idx0< (zrel.size-1)):       # point below water
        #    d0 = np.mean(d[idx0:idx0+2])
        #    cd0 = np.mean(cd[idx0:idx0+2])
        # elif (zrel[idx0]>0.) and (idx0>0):     # point above water
        #    d0 = np.mean(d[idx0-1:idx0+1])
        #    cd0 = np.mean(cd[idx0-1:idx0+1])
        # Fi0 = rho*inputs['cm']*math.pi/4.0*d0**2*inputs['A0']  # Morrison's equation
        # Fd0 = q0*cd0*d0
        # Fp0 = Fi0 + Fd0

        # Px0 = Fp0*cosd(beta0)
        # Py0 = Fp0*sind(beta0)
        # Pz0 = 0.*Fp0

        # Store qties at z=0 MSL
        # outputs['waveLoads_Px0'] = Px0
        # outputs['waveLoads_Py0'] = Py0
        # outputs['waveLoads_Pz0'] = Pz0
        # outputs['waveLoads_qdyn0'] = q0
        # outputs['waveLoads_beta0'] = beta0

        # pack data
        outputs["waveLoads_Px"] = Px
        outputs["waveLoads_Py"] = Py
        outputs["waveLoads_Pz"] = Pz
        outputs["waveLoads_qdyn"] = q
        outputs["waveLoads_pt"] = q + inputs["p"]
        outputs["waveLoads_z"] = inputs["z"]
        outputs["waveLoads_beta"] = beta

    def compute_partials(self, inputs, J):

        # wlevel = inputs['wlevel']
        # if wlevel > 0.0: wlevel *= -1.0

        rho = inputs["rho_water"]
        U = inputs["U"]
        # U0 = inputs['U0']
        d = inputs["d"]
        # zrel= inputs['z']-wlevel
        mu = inputs["mu_water"]
        beta = inputs["beta_wave"]
        # beta0 = inputs['beta0']

        # dynamic pressure
        q = 0.5 * rho * U ** 2
        # q0= 0.5*rho*U0**2

        # Reynolds number and drag
        if float(inputs["cd_usr"]) < 0.0:
            cd = inputs["cd_usr"] * np.ones_like(d)
            Re = 1.0
            dcd_dRe = 0.0
        else:
            Re = rho * U * d / mu
            cd, dcd_dRe = cylinderDrag(Re)

        # derivatives
        dq_dU = rho * U
        const = (dq_dU * cd + q * dcd_dRe * rho * d / mu) * d
        dPx_dU = const * cosd(beta)
        dPy_dU = const * sind(beta)

        const = (cd + dcd_dRe * Re) * q + rho * inputs["cm"] * math.pi / 4.0 * 2 * d * inputs["A"]
        dPx_dd = const * cosd(beta)
        dPy_dd = const * sind(beta)

        const = rho * inputs["cm"] * math.pi / 4.0 * d ** 2
        dPx_dA = const * cosd(beta)
        dPy_dA = const * sind(beta)

        J["waveLoads_Px", "A"] = dPx_dA
        J["waveLoads_Py", "A"] = dPy_dA
        J["waveLoads_qdyn", "U"] = dq_dU
        J["waveLoads_pt", "U"] = dq_dU


# ___________________________________________#


class CylinderEnvironment(om.Group):
    def initialize(self):
        self.options.declare("wind", default="power")
        self.options.declare("nPoints")
        self.options.declare("water_flag", default=True)

    def setup(self):
        nPoints = self.options["nPoints"]
        wind = self.options["wind"]
        water_flag = self.options["water_flag"]

        self.set_input_defaults("z0", 0.0)
        self.set_input_defaults("cd_usr", -1.0)
        self.set_input_defaults("yaw", 0.0, units="deg")

        self.set_input_defaults("beta_wind", 0.0, units="deg")
        self.set_input_defaults("rho_air", 1.225, units="kg/m**3")
        self.set_input_defaults("mu_air", 1.81206e-5, units="kg/m/s")
        self.set_input_defaults("shearExp", 0.2)

        if water_flag:
            self.set_input_defaults("beta_wave", 0.0, units="deg")
            self.set_input_defaults("rho_water", 1025.0, units="kg/m**3")
            self.set_input_defaults("mu_water", 1.08e-3, units="kg/m/s")

        # Wind profile and loads
        promwind = ["Uref", "zref", "z", "z0"]
        if wind is None or wind.lower() in ["power", "powerwind", ""]:
            self.add_subsystem("wind", PowerWind(nPoints=nPoints), promotes=promwind + ["shearExp"])

        elif wind.lower() == "logwind":
            self.add_subsystem("wind", LogWind(nPoints=nPoints), promotes=promwind)

        else:
            raise ValueError("Unknown wind type, " + wind)

        self.add_subsystem(
            "windLoads",
            CylinderWindDrag(nPoints=nPoints),
            promotes=["cd_usr", "beta_wind", "rho_air", "mu_air", "z", "d"],
        )

        # Wave profile and loads
        if water_flag:
            self.add_subsystem(
                "wave",
                LinearWaves(nPoints=nPoints),
                promotes=[
                    "z",
                    "Uc",
                    "Hsig_wave",
                    "Tsig_wave",
                    "rho_water",
                    ("z_floor", "water_depth"),
                    ("z_surface", "z0"),
                ],
            )

            self.add_subsystem(
                "waveLoads",
                CylinderWaveDrag(nPoints=nPoints),
                promotes=["cm", "cd_usr", "beta_wave", "rho_water", "mu_water", "z", "d"],
            )

        # Combine all loads
        self.add_subsystem(
            "distLoads", AeroHydroLoads(nPoints=nPoints), promotes=["Px", "Py", "Pz", "qdyn", "yaw", "z"]
        )

        # Connections
        self.connect("wind.U", "windLoads.U")
        if water_flag:
            self.connect("wave.U", "waveLoads.U")
            self.connect("wave.A", "waveLoads.A")
            self.connect("wave.p", "waveLoads.p")

        self.connect("windLoads.windLoads_Px", "distLoads.windLoads_Px")
        self.connect("windLoads.windLoads_Py", "distLoads.windLoads_Py")
        self.connect("windLoads.windLoads_Pz", "distLoads.windLoads_Pz")
        self.connect("windLoads.windLoads_qdyn", "distLoads.windLoads_qdyn")
        self.connect("windLoads.windLoads_beta", "distLoads.windLoads_beta")
        self.connect("windLoads.windLoads_z", "distLoads.windLoads_z")

        if water_flag:
            self.connect("waveLoads.waveLoads_Px", "distLoads.waveLoads_Px")
            self.connect("waveLoads.waveLoads_Py", "distLoads.waveLoads_Py")
            self.connect("waveLoads.waveLoads_Pz", "distLoads.waveLoads_Pz")
            self.connect("waveLoads.waveLoads_pt", "distLoads.waveLoads_qdyn")
            self.connect("waveLoads.waveLoads_beta", "distLoads.waveLoads_beta")
            self.connect("waveLoads.waveLoads_z", "distLoads.waveLoads_z")


def main():
    # initialize problem
    U = np.array([20.0, 25.0, 30.0])
    z = np.array([10.0, 30.0, 80.0])
    d = np.array([5.5, 4.0, 3.0])

    beta = np.array([45.0, 45.0, 45.0])
    rho = 1.225
    mu = 1.7934e-5
    # cd_usr = 0.7

    nPoints = len(z)

    prob = om.Problem()

    root = prob.model = om.Group()

    root.add("p1", CylinderWindDrag(nPoints))

    prob.setup()

    prob["p1.U"] = U
    prob["p1.z"] = z
    prob["p1.d"] = d
    prob["p1.beta"] = beta
    prob["p1.rho"] = rho
    prob["p1.mu"] = mu
    # prob['p1.cd_usr'] = cd_usr

    # run
    prob.run_once()

    # out
    Re = prob["p1.rho"] * prob["p1.U"] * prob["p1.d"] / prob["p1.mu"]
    cd, dcd_dRe = cylinderDrag(Re)
    print(cd)
    import matplotlib.pyplot as plt

    plt.plot(prob["p1.windLoads_Px"], prob["p1.windLoads_z"])
    plt.plot(prob["p1.windLoads_Py"], prob["p1.windLoads_z"])
    plt.plot(prob["p1.windLoads_qdyn"], prob["p1.windLoads_z"])
    plt.show()


if __name__ == "__main__":
    main()



"""Common processes and cargo types for quayside assembly and tow-out
installations"""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


from marmot import Agent, le, process
from marmot._exceptions import AgentNotRegistered


class Substructure:
    """Floating Substructure Class."""

    def __init__(self):
        """Creates an instance of `Substructure`."""

        pass


class SubstructureAssemblyLine(Agent):
    """Substructure Assembly Line Class."""

    def __init__(self, assigned, time, target, num):
        """
        Creates an instance of `SubstructureAssemblyLine`.

        Parameters
        ----------
        assigned : list
            List of assigned tasks. Can be shared with other assembly lines.
        time : int | float
            Hours required to produce one substructure.
        target : simpy.Store
            Target storage.
        num : int
            Assembly line number designation.
        """

        super().__init__(f"Substructure Assembly Line {num}")

        self.assigned = assigned
        self.time = time
        self.target = target

    def submit_action_log(self, action, duration, **kwargs):
        """
        Submits a log representing a completed `action` performed over time
        `duration`.

        This method overwrites the default `submit_action_log` in
        `marmot.Agent`, adding operation cost to every submitted log within
        ORBIT.

        Parameters
        ----------
        action : str
            Performed action.
        duration : int | float
            Duration of action.

        Raises
        ------
        AgentNotRegistered
        """

        if self.env is None:
            raise AgentNotRegistered(self)

        else:
            payload = {
                **kwargs,
                "agent": str(self),
                "action": action,
                "duration": float(duration),
                "cost": 0,
            }

            self.env._submit_log(payload, level="ACTION")

    @process
    def assemble_substructure(self):
        """
        Simulation process for assembling a substructure.
        """

        yield self.task("Substructure Assembly", self.time)
        substructure = Substructure()

        start = self.env.now
        yield self.target.put(substructure)
        delay = self.env.now - start

        if delay > 0:
            self.submit_action_log("Delay: No Wet Storage Available", delay)

    @process
    def start(self):
        """
        Trigger the assembly line to run. Will attempt to pull a task from
        self.assigned and timeout for the assembly time. Shuts down after
        self.assigned is empty.
        """

        while True:
            try:
                _ = self.assigned.pop(0)
                yield self.assemble_substructure()

            except IndexError:
                break


class TurbineAssemblyLine(Agent):
    """Turbine Assembly Line Class."""

    def __init__(self, feed, target, turbine, num):
        """
        Creates an instance of `TurbineAssemblyLine`.

        Parameters
        ----------
        feed : simpy.Store
            Storage for completed substructures.
        target : simpy.Store
            Target storage.
        num : int
            Assembly line number designation.
        """

        super().__init__(f"Turbine Assembly Line {num}")

        self.feed = feed
        self.target = target
        self.turbine = turbine

    def submit_action_log(self, action, duration, **kwargs):
        """
        Submits a log representing a completed `action` performed over time
        `duration`.

        This method overwrites the default `submit_action_log` in
        `marmot.Agent`, adding operation cost to every submitted log within
        ORBIT.

        Parameters
        ----------
        action : str
            Performed action.
        duration : int | float
            Duration of action.

        Raises
        ------
        AgentNotRegistered
        """

        if self.env is None:
            raise AgentNotRegistered(self)

        else:
            payload = {
                **kwargs,
                "agent": str(self),
                "action": action,
                "duration": float(duration),
                "cost": 0,
            }

            self.env._submit_log(payload, level="ACTION")

    @process
    def start(self):
        """
        Trigger the assembly line to run. Will attempt to pull a task from
        self.assigned and timeout for the assembly time. Shuts down after
        self.assigned is empty.
        """

        while True:
            start = self.env.now
            sub = yield self.feed.get()
            delay = self.env.now - start

            if delay > 0:
                self.submit_action_log("Delay: No Substructures in Wet Storage", delay)

            yield self.assemble_turbine()

    @process
    def assemble_turbine(self):
        """
        Turbine assembly process. Follows a similar process as the
        `TurbineInstallation` modules but has fixed lift times + fasten times
        instead of calculating the lift times dynamically.
        """

        yield self.move_substructure()
        yield self.prepare_for_assembly()

        sections = self.turbine["tower"].get("sections", 1)
        for _ in range(sections):
            yield self.lift_and_attach_tower_section()

        yield self.lift_and_attach_nacelle()

        for _ in range(3):
            yield self.lift_and_attach_blade()

        yield self.mechanical_completion()

        start = self.env.now
        yield self.target.put(1)
        delay = self.env.now - start

        if delay > 0:
            self.submit_action_log("Delay: No Assembly Storage Available", delay)

        self.submit_debug_log(message="Assembly delievered to installation groups.")

    @process
    def move_substructure(self):
        """
        Task representing time associated with moving the completed
        substructure assembly to the turbine assembly line.

        TODO: Move to dynamic process involving tow groups.
        """

        yield self.task("Move Substructure", 8)

    @process
    def prepare_for_assembly(self):
        """
        Task representing time associated with preparing a substructure for
        turbine assembly.
        """

        yield self.task("Prepare for Turbine Assembly", 12)

    @process
    def lift_and_attach_tower_section(self):
        """
        Task representing time associated with lifting and attaching a tower
        section at quayside.
        """

        yield self.task(
            "Lift and Attach Tower Section",
            12,
            constraints={"windspeed": le(15)},
        )

    @process
    def lift_and_attach_nacelle(self):
        """
        Task representing time associated with lifting and attaching a nacelle
        at quayside.
        """

        yield self.task("Lift and Attach Nacelle", 7, constraints={"windspeed": le(15)})

    @process
    def lift_and_attach_blade(self):
        """
        Task representing time associated with lifting and attaching a turbine
        blade at quayside.
        """

        yield self.task("Lift and Attach Blade", 3.5, constraints={"windspeed": le(12)})

    @process
    def mechanical_completion(self):
        """
        Task representing time associated with performing mechanical compltion
        work at quayside.
        """

        yield self.task("Mechanical Completion", 24, constraints={"windspeed": le(18)})


class TowingGroup(Agent):
    """Class to represent an arbitrary group of towing vessels."""

    def __init__(self, vessel_specs, num=1):
        """
        Creates an instance of TowingGroup.

        Parameters
        ----------
        vessel_specs : dict
            Specs for the individual vessels used in the towing group.
            Currently restricted to one vessel specification per group.
        """

        super().__init__(f"Towing Group {num}")
        self._specs = vessel_specs
        self.day_rate = self._specs["vessel_specs"]["day_rate"]
        self.transit_speed = self._specs["transport_specs"]["transit_speed"]

    def initialize(self):
        """Initializes the towing group."""

        self.submit_debug_log(message="{self.name} initialized.")

    @process
    def group_task(self, name, duration, num_vessels, constraints={}, **kwargs):
        """
        Submits a group task with any number of towing vessels.

        Parameters
        ----------
        name : str
            Name of task to complete. Used for submitting action logs.
        duration : float | int
            Duration of the task.
            Rounded up to the nearest int.
        num_vessels : int
            Number of individual towing vessels needed for the operation.
        """

        kwargs = {**kwargs, "num_vessels": num_vessels}
        yield self.task(name, duration, constraints=constraints, **kwargs)

    def operation_cost(self, hours, **kwargs):
        """
        Returns cost of an operation of duration `hours` using number of
        vessels, `num_vessels`.

        Parameters
        ----------
        hours : int | float
            Duration of operation in hours.
        vessels : int
            Default: 1
        """

        mult = kwargs.get("cost_multiplier", 1.0)
        vessels = kwargs.get("num_vessels", 1)
        return (self.day_rate / 24) * vessels * hours * mult

    def submit_action_log(self, action, duration, **kwargs):
        """
        Submits a log representing a completed `action` performed over time
        `duration`.

        This method overwrites the default `submit_action_log` in
        `marmot.Agent`, adding operation cost to every submitted log within
        ORBIT.

        Parameters
        ----------
        action : str
            Performed action.
        duration : int | float
            Duration of action.

        Raises
        ------
        AgentNotRegistered
        """

        if self.env is None:
            raise AgentNotRegistered(self)

        else:
            payload = {
                **kwargs,
                "agent": str(self),
                "action": action,
                "duration": float(duration),
                "cost": self.operation_cost(duration, **kwargs),
            }

            self.env._submit_log(payload, level="ACTION")



import os
import unittest

import openmdao.api as om
from wisdem.rotorse.blade_cost import StandaloneBladeCost, initialize_omdao_prob
from wisdem.glue_code.gc_LoadInputs import WindTurbineOntologyPython
from wisdem.glue_code.gc_PoseOptimization import PoseOptimization

wisdem_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__)))))
example_dir02 = os.path.join(wisdem_dir, "examples", "02_reference_turbines")  # get path example 02_reference_turbines
example_dir03 = os.path.join(wisdem_dir, "examples", "03_blade")  # get path example 03_blade
fname_opt_options = os.path.join(example_dir02, "analysis_options.yaml")
accuracy = 0


class TestBC(unittest.TestCase):
    def testBladeCostIEA3p4(self):

        fname_modeling_options = os.path.join(example_dir02, "modeling_options.yaml")
        fname_wt_input = os.path.join(example_dir02, "IEA-3p4-130-RWT.yaml")
        wt_initial = WindTurbineOntologyPython(fname_wt_input, fname_modeling_options, fname_opt_options)
        wt_init, modeling_options, opt_options = wt_initial.get_input_data()
        modeling_options["WISDEM"]["RotorSE"]["flag"] = False
        wt_opt = om.Problem(model=StandaloneBladeCost(modeling_options=modeling_options, opt_options=opt_options))
        wt_opt.setup(derivatives=False)
        myopt = PoseOptimization(wt_init, modeling_options, opt_options)
        wt_opt = myopt.set_initial(wt_opt, wt_init)
        wt_opt = initialize_omdao_prob(wt_opt, modeling_options, wt_init)
        wt_opt.run_model()

        self.assertAlmostEqual(wt_opt["rc.total_labor_hours"][0], 950.108886892848, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_non_gating_ct"][0], 136.09304978006895, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_metallic_parts_cost"][0], 4667.410277624349, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_consumable_cost_w_waste"][0], 6893.931399813539, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_blade_mat_cost_w_waste"][0], 69437.29121589664, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_cost_labor"][0], 32059.8612441465, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_cost_utility"][0], 647.567711197268, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.blade_variable_cost"][0], 102144.7201712404, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_cost_equipment"][0], 3543.7049227917837, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_cost_tooling"][0], 9764.710077639173, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_cost_building"][0], 572.1105024592529, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_maintenance_cost"][0], 3666.368184490084, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_labor_overhead"][0], 9617.95837324395, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.cost_capital"][0], 7363.293915205925, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.blade_fixed_cost"][0], 34528.145975830164, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_blade_cost"][0], 136672.86614707057, places=accuracy)

    def testBladeCostIEA10(self):

        fname_modeling_options = os.path.join(example_dir02, "modeling_options_iea10.yaml")
        fname_wt_input = os.path.join(example_dir02, "IEA-10-198-RWT.yaml")
        wt_initial = WindTurbineOntologyPython(fname_wt_input, fname_modeling_options, fname_opt_options)
        wt_init, modeling_options, opt_options = wt_initial.get_input_data()
        modeling_options["WISDEM"]["RotorSE"]["flag"] = False
        modeling_options["WISDEM"]["DriveSE"]["direct"] = False
        wt_opt = om.Problem(model=StandaloneBladeCost(modeling_options=modeling_options, opt_options=opt_options))
        wt_opt.setup(derivatives=False)
        myopt = PoseOptimization(wt_init, modeling_options, opt_options)
        wt_opt = myopt.set_initial(wt_opt, wt_init)
        wt_opt = initialize_omdao_prob(wt_opt, modeling_options, wt_init)
        wt_opt.run_model()

        self.assertAlmostEqual(wt_opt["rc.total_labor_hours"][0], 1984.554250372168, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_non_gating_ct"][0], 229.3933649297149, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_metallic_parts_cost"][0], 7621.331066101973, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_consumable_cost_w_waste"][0], 12200.172713827997, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_blade_mat_cost_w_waste"][0], 195262.04234970373, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_cost_labor"][0], 67080.47848122781, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_cost_utility"][0], 2247.2660503124343, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.blade_variable_cost"][0], 264589.786881244, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_cost_equipment"][0], 8679.713083079712, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_cost_tooling"][0], 31598.877474314024, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_cost_building"][0], 1925.1278368386509, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_maintenance_cost"][0], 10837.859033328512, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_labor_overhead"][0], 20124.14354436834, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.cost_capital"][0], 21050.71145769286, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.blade_fixed_cost"][0], 94216.43242962209, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_blade_cost"][0], 358806.2193108661, places=accuracy)

    def testBladeCostIEA15(self):

        fname_modeling_options = os.path.join(example_dir02, "modeling_options.yaml")
        fname_wt_input = os.path.join(example_dir02, "IEA-15-240-RWT.yaml")
        wt_initial = WindTurbineOntologyPython(fname_wt_input, fname_modeling_options, fname_opt_options)
        wt_init, modeling_options, opt_options = wt_initial.get_input_data()
        modeling_options["WISDEM"]["RotorSE"]["flag"] = False
        modeling_options["WISDEM"]["DriveSE"]["direct"] = False
        wt_opt = om.Problem(model=StandaloneBladeCost(modeling_options=modeling_options, opt_options=opt_options))
        wt_opt.setup(derivatives=False)
        myopt = PoseOptimization(wt_init, modeling_options, opt_options)
        wt_opt = myopt.set_initial(wt_opt, wt_init)
        wt_opt = initialize_omdao_prob(wt_opt, modeling_options, wt_init)
        wt_opt.run_model()

        self.assertAlmostEqual(wt_opt["rc.total_labor_hours"][0], 3133.7875857844165, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_non_gating_ct"][0], 276.2981847459007, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_metallic_parts_cost"][0], 8885.387764565836, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_consumable_cost_w_waste"][0], 15045.283018454009, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_blade_mat_cost_w_waste"][0], 361485.95863826084, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_cost_labor"][0], 105876.57752725758, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_cost_utility"][0], 3628.3329952322097, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.blade_variable_cost"][0], 470990.86916075065, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_cost_equipment"][0], 25795.210670398752, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_cost_tooling"][0], 32212.11155993386, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_cost_building"][0], 2369.9813509855767, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_maintenance_cost"][0], 18315.99973893161, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_labor_overhead"][0], 31762.973258177273, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.cost_capital"][0], 35940.02567060387, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.blade_fixed_cost"][0], 146396.30224903094, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc.total_blade_cost"][0], 617387.1714097816, places=accuracy)

    def testBladeCostBAR_USC(self):

        fname_modeling_options = os.path.join(example_dir03, "modeling_options.yaml")
        fname_wt_input = os.path.join(example_dir03, "BAR_USC.yaml")
        wt_initial = WindTurbineOntologyPython(fname_wt_input, fname_modeling_options, fname_opt_options)
        wt_init, modeling_options, opt_options = wt_initial.get_input_data()
        modeling_options["WISDEM"]["RotorSE"]["flag"] = False
        wt_opt = om.Problem(model=StandaloneBladeCost(modeling_options=modeling_options, opt_options=opt_options))
        wt_opt.setup(derivatives=False)
        myopt = PoseOptimization(wt_init, modeling_options, opt_options)
        wt_opt = myopt.set_initial(wt_opt, wt_init)
        wt_opt = initialize_omdao_prob(wt_opt, modeling_options, wt_init)
        wt_opt.run_model()

        self.assertAlmostEqual(wt_opt["rc_in.total_labor_hours"][0], 1773.5545744866824, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc_in.total_non_gating_ct"][0], 196.0218483852534, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc_in.total_metallic_parts_cost"][0], 6482.0815222645215, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc_in.total_consumable_cost_w_waste"][0], 8807.09170136831, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc_in.total_blade_mat_cost_w_waste"][0], 280634.2143258618, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc_in.total_cost_labor"][0], 59930.03202124485, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc_in.total_cost_utility"][0], 1852.7768421527749, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc_in.blade_variable_cost"][0], 342417.0231892594, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc_in.total_cost_equipment"][0], 10146.9503312002, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc_in.total_cost_tooling"][0], 17156.69264865687, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc_in.total_cost_building"][0], 845.6837995649553, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc_in.total_maintenance_cost"][0], 7818.671515743128, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc_in.total_labor_overhead"][0], 17979.009606373453, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc_in.cost_capital"][0], 17654.04402081108, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc_in.blade_fixed_cost"][0], 71601.05192234968, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc_in.total_blade_cost"][0], 414018.07511160907, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc_out.total_labor_hours"][0], 342.7617494900632, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc_out.total_non_gating_ct"][0], 69.2156025516638, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc_out.total_metallic_parts_cost"][0], 3428.7667570837384, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc_out.total_consumable_cost_w_waste"][0], 2898.312583402607, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc_out.total_blade_mat_cost_w_waste"][0], 30334.814048848515, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc_out.total_cost_labor"][0], 11539.108916022531, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc_out.total_cost_utility"][0], 166.05676723016776, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc_out.blade_variable_cost"][0], 42039.979732101216, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc_out.total_cost_equipment"][0], 1416.8953538006472, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc_out.total_cost_tooling"][0], 2510.9560206969245, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc_out.total_cost_building"][0], 177.15501917770482, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc_out.total_maintenance_cost"][0], 1181.097127845013, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc_out.total_labor_overhead"][0], 3461.7326748067594, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc_out.cost_capital"][0], 2538.6441396252812, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc_out.blade_fixed_cost"][0], 11286.48033595233, places=accuracy)
        self.assertAlmostEqual(wt_opt["rc_out.total_blade_cost"][0], 53326.460068053544, places=accuracy)
        self.assertAlmostEqual(wt_opt["total_bc.total_blade_cost"][0], 485944.53517966263, places=accuracy)

def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestBC))
    return suite


if __name__ == "__main__":
    result = unittest.TextTestRunner().run(suite())

    if result.wasSuccessful():
        exit(0)
    else:
        exit(1)



import numpy as np
from wisdem.commonse import eps

# Compute costs based on "Optimum Design of Steel Structures" by Farkas and Jarmai
# All correlations are based in mm and all inputs are assumed to be in m and then converted within the functions


def steel_cutting_plasma_time(length, thickness):
    # Length input as meters, thickness in mm
    # time = length / (-0.180150943 + 41.03815215/(1e3*thickness+eps)) # minutes
    # Better conditioned polynomial fit to the above correlation
    pp = np.array([2.44908121e02, 1.74461814e01, 7.05214799e-02])
    time = length * np.polyval(pp, thickness)
    return np.sum(time)


def steel_rolling_time(theta, radius, thickness):
    # Radius and thickness input as meters, converted to mm
    time = theta * np.exp(
        6.8582513 - 4.527217 / np.sqrt(1e3 * thickness + eps) + 0.009541996 * np.sqrt(2 * 1e3 * radius)
    )
    return np.sum(time)


def steel_tube_cutgrind_time(theta, radius, thickness, angle):
    # Radius and thickness input as meters, converted to mm
    # time = theta * 2.5 * np.pi * (2.0*1e3*radius) / ((350.0 - 2.0*1e3*thickness)*0.3*np.sin(angle))
    # Better conditioned polynomial fit to the above correlation
    pp = np.array([7.84235859, 0.1428632, 0.07765389])
    time = theta * (2.0 * 1e3 * radius) / (np.polyval(pp, thickness) * np.sin(angle))
    return np.sum(time)


def steel_welding_time(theta, npieces, mtotal, length, thickness, coeff):
    # Length input as meters, thickness as mm
    time = np.sum(theta * np.sqrt(npieces * mtotal))
    time += np.sum(1.3e-3 * coeff * (length) * (1e3 * thickness) ** 1.9358)
    return time


def steel_butt_welding_time(theta, npieces, mtotal, length, thickness):
    return steel_welding_time(theta, npieces, mtotal, length, thickness, 0.152)


def steel_filett_welding_time(theta, npieces, mtotal, length, thickness):
    return steel_welding_time(theta, npieces, mtotal, length, thickness, 0.3394)


def steel_tube_welding_time(theta, npieces, mtotal, length, thickness):
    return steel_welding_time(theta, npieces, mtotal, length, thickness, 0.7889)



"""
Provides a library for sub configurations within ORBIT. By default, the library
is located at `../ORBIT/library/`, but it can also be stored outside of the
repo by passing in the `library_path` variable to `ProjectManager` or any
individual phases.

ORBIT expects a library outside of the repo to have the following structure:
```
<library folder name>
├── defaults         <- Top-level default data
├── project
│   ├── config       <- Configuration dictionary repository
│   ├── port         <- Port specific data setttings
│   ├── plant        <- Wind farm specific data setttings
│   ├── site         <- Project site data settings
│   ├── development  <- Project development cost settings
├── cables           <- Cable data files: array cables, export cables
├── substructures    <- Substructure data files: monopiles, jackets, etc.
├── turbines         <- Turbine data files
├── vessels          <- Vessel data files
│   ├── defaults     <- Default data related to vessel tasks
├── weather          <- Weather profiles
├── results
```
"""

__author__ = "Rob Hammond"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Rob Hammond"
__email__ = "robert.hammond@nrel.gov"


import os
import re
import csv
import warnings

import yaml
import pandas as pd
from yaml import Dumper

from wisdem.orbit.core.exceptions import LibraryItemNotFoundError

ROOT = os.path.abspath(os.path.join(os.path.abspath(__file__), "../../.."))
default_library = os.path.join(ROOT, "library")

# Need a custom loader to read in scientific notation correctly
class CustomSafeLoader(yaml.SafeLoader):
    def construct_python_tuple(self, node):
        return tuple(self.construct_sequence(node))


CustomSafeLoader.add_constructor("tag:yaml.org,2002:python/tuple", CustomSafeLoader.construct_python_tuple)

loader = CustomSafeLoader
loader.add_implicit_resolver(
    "tag:yaml.org,2002:float",
    re.compile(
        """^(?:
     [-+]?(?:[0-9][0-9_]*)\\.[0-9_]*(?:[eE][-+]?[0-9]+)?
    |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
    |\\.[0-9_]+(?:[eE][-+][0-9]+)?
    |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\\.[0-9_]*
    |[-+]?\\.(?:inf|Inf|INF)
    |\\.(?:nan|NaN|NAN))$""",
        re.X,
    ),
    list("-+0123456789."),
)


def clean_warning(message, category, filename, lineno, file=None, line=""):
    """Formats the standard warning output."""
    return f"{category.__name__}: {filename}:{lineno}\n{message}"


warnings.formatwarning = clean_warning


def initialize_library(library_path):
    """
    Creates an environment variable named "DATA_LIBRARY", defaulting to
    `../ORBIT/library/` if not defined.

    Parameters
    ----------
    library_path : str | None
        Absolute path to the project library.
    """

    if "DATA_LIBRARY" in os.environ:
        return

    if library_path is None:
        library_path = default_library

    if not os.path.isdir(library_path):
        raise ValueError(f"Invalid library path.")

    os.environ["DATA_LIBRARY"] = library_path
    print(f"ORBIT library intialized at '{library_path}'")


def extract_library_data(config, additional_keys=[]):
    """
    Extracts the configuration data from the specified library.

    Parameters
    ----------
    config : dict
        Configuration dictionary.
    additional_keys : list
        Additional keys that contain data that needs to be extracted from
        within `config`, by default [].

    Returns
    -------
    config : dict
        Configuration dictionary.
    """

    if os.environ.get("DATA_LIBRARY", None) is None:
        return config

    for key, val in config.items():
        if isinstance(val, dict) and any(el in key for el in additional_keys):
            config[key] = extract_library_data(val)
        elif not isinstance(val, str):
            continue

        try:
            config[key] = extract_library_specs(key, val)
        except KeyError:
            continue

    return config


def extract_library_specs(key, filename, file_type="yaml"):
    """
    Base method that extracts a file from the configured data library. If the
    file is not found in the configured library, the default library located
    within the repository will also be searched.

    Parameters
    ----------
    key : str
        Configuration key used to map to specific subpath in library.
    filename : str
        Name of the file to be extracted.
    file_type : str
        Should be one of "yaml" or "csv".

    Returns
    -------
    dict
        Dictionary of specifications for `filename`.

    Raises
    ------
    LibraryItemNotFoundError
        An error is raised when the file cannot be found in the library or the
        default library.
    """

    filename = f"{filename}.{file_type}"
    path = PATH_LIBRARY[key]
    filepath = os.path.join(os.environ["DATA_LIBRARY"], path, filename)

    if os.path.isfile(filepath):
        return _extract_file(filepath)

    if os.environ["DATA_LIBRARY"] != default_library:
        filepath = os.path.join(default_library, path, filename)
        if os.path.isfile(filepath):
            return _extract_file(filepath)

    raise LibraryItemNotFoundError(path, filename)


def _extract_file(filepath):
    """
    Extracts file from valid filepath. Currently only supports "yaml" or "csv".

    Parameters
    ----------
    filepath : str
        Valid filepath of library item.
    """

    if filepath.endswith("yaml"):
        f = open(filepath, "r")
        fyaml = yaml.load(f, Loader=loader)
        f.close()
        return fyaml

    elif filepath.endswith("csv"):
        df = pd.read_csv(filepath, index_col=False)

        # Drop empty rows and columns
        df.dropna(how="all", inplace=True)
        df.dropna(how="all", inplace=True, axis=1)

        # Enforce strictly lowercase and "_" separated column names
        df.columns = [el.replace(" ", "_").lower() for el in df.columns]
        return df

    else:
        _type = filepath.split(".")[-1]
        raise TypeError(f"File type {_type} not supported for extraction.")


def _get_yes_no_response(filename):
    """Elicits a y/n response from the user to overwrite a file.

    Returns
    -------
    bool
        Indicator to overwrite `filename`.
    """

    response = input(f"{filename} already exists, overwrite [y/n]?").lower()
    if response not in ("y", "n"):
        print("Bad input! Must be one of [y/n]")
        _get_yes_no_response(filename)
    return True if response == "y" else False


def export_library_specs(key, filename, data, file_ext="yaml"):
    """
    Base method that export a file to the data library.

    Parameters
    ----------
    key : str
        Configuration key used to map to a specific subpath in library.
    filename : str
        Name to be given to the file without the extension.
    data : yaml-ready or List[list]
        Data to be saved to YAML (any Python type) or csv-ready.
    """

    filename = f"{filename}.{file_ext}"
    path = PATH_LIBRARY[key]
    data_path = os.path.join(os.environ["DATA_LIBRARY"], path, filename)
    if os.path.isfile(data_path) and not _get_yes_no_response(data_path):
        print("Cancelling save!")
        return
    if file_ext == "yaml":
        f = open(data_path, "w")
        yaml.dump(data, f, Dumper=Dumper, default_flow_style=False)
        f.close()
    elif file_ext == "csv":
        with open(data_path, "w") as f:
            writer = csv.writer(f)
            writer.writerows(data)
    print("Save complete!")


PATH_LIBRARY = {
    # default data
    "defaults": "defaults",
    # vessels
    "array_cable_install_vessel": "vessels",
    "array_cable_bury_vessel": "vessels",
    "array_cable_trench_vessel": "vessels",
    "export_cable_install_vessel": "vessels",
    "export_cable_bury_vessel": "vessels",
    "export_cable_trench_vessel": "vessels",
    "oss_install_vessel": "vessels",
    "spi_vessel": "vessels",
    "trench_dig_vessel": "vessels",
    "feeder": "vessels",
    "mooring_install_vessel": "vessels",
    "wtiv": "vessels",
    "towing_vessel": "vessels",
    "support_vessel": "vessels",
    # cables
    "cables": "cables",
    "array_system": "cables",
    "array_system_design": "cables",
    "export_system": "cables",
    "export_system_design": "cables",
    # project details
    "config": os.path.join("project", "config"),
    "plant": os.path.join("project", "plant"),
    "port": os.path.join("project", "ports"),
    "project_development": os.path.join("project", "development"),
    "site": os.path.join("project", "site"),
    # substructures
    "monopile": "substructures",
    "monopile_design": "substructures",
    "scour_protection": "substructures",
    "scour_design": "substructures",
    "transition_piece": "substructures",
    "offshore_substation_substructure": "substructures",
    "offshore_substation_topside": "substructures",
    # turbine
    "turbine": "turbines",
}



from __future__ import print_function

import abc

import numpy as np


class CrossSectionBase(metaclass=abc.ABCMeta):
    @property
    @abc.abstractmethod
    def Area(self):
        pass

    @property
    @abc.abstractmethod
    def Ixx(self):  # 2nd area moment of inertia w.r.t. x-x axis
        pass

    @property
    @abc.abstractmethod
    def Iyy(self):  # 2nd area moment of inertia w.r.t. z-z axis
        pass

    @property
    @abc.abstractmethod
    def J0(self):  # polar moment of inertia w.r.t. z-z axis (torsional)
        pass

    @property
    @abc.abstractmethod
    def Asx(self):  # Shear Area
        pass

    @property
    @abc.abstractmethod
    def Asy(self):  # Shear Area
        pass

    @property
    @abc.abstractmethod
    def BdgMxx(self):  # Bending modulus
        pass

    @property
    @abc.abstractmethod
    def BdgMyy(self):  # Bending modulus
        pass

    @property
    @abc.abstractmethod
    def TorsConst(self):  # Torsion shear constant
        pass

    @property
    def Sx(self):  # Bending modulus
        return self.BdgMxx

    @property
    def Sy(self):  # Bending modulus
        return self.BdgMyy

    @property
    def C(self):  # Torsion shear constant
        return self.TorsConst


class Tube(CrossSectionBase):
    """The Tube Class contains functions to calculate properties of tubular circular cross-sections
    for structural analyses."""

    def __init__(self, D, t, L=np.NaN, Kbuck=1.0):
        self.D = D
        self.R = 0.5 * D
        self.t = t
        self.L = L * np.ones(np.size(D))
        self.Kbuck = Kbuck * np.ones(np.size(D))

    @property
    def Area(self):  # Cross sectional area of tube
        return (self.D ** 2 - (self.D - 2 * self.t) ** 2) * np.pi / 4.0

    @property
    def derivArea(self):
        return {"D": 2 * np.pi / 4 * (self.D ** 2 - (self.D - 2 * self.t)) * (2 * self.D - 1), "t": 0}

    @property
    def Amid(self):  # mid-thickness inscribed area of tube (thin wall torsion calculation)
        return (self.D - self.t) ** 2 * np.pi / 4.0

    @property
    def Ixx(self):  # 2nd area moment of inertia w.r.t. x-x axis (Iyy=Ixx for tube)
        return (self.D ** 4 - (self.D - 2 * self.t) ** 4) * np.pi / 64.0

    @property
    def Iyy(self):  # 2nd area moment of inertia w.r.t. z-z axis (Iyy=Ixx for tube)
        return self.Ixx

    @property
    def J0(self):  # polar moment of inertia w.r.t. z-z axis (torsional)
        return 2.0 * self.Ixx

    @property
    def Asx(self):  # Shear Area for tubular cross-section
        Ri = self.D / 2 - self.t
        Ro = self.R
        return self.Area / (1.124235 + 0.055610 * (Ri / Ro) + 1.097134 * (Ri / Ro) ** 2 - 0.630057 * (Ri / Ro) ** 3)

    @property
    def Asy(self):  # Shear Area for tubular cross-section
        return self.Asx

    @property
    def BdgMxx(self):  # Bending modulus for tubular cross-section
        return self.Ixx / self.R

    @property
    def BdgMyy(self):  # Bending modulus for tubular cross-section =BdgMxx
        return self.Iyy / self.R

    @property
    def TorsConst(self):  # Torsion shear constant for tubular cross-section
        return self.J0 / self.R

    @property
    def S(self):  # Bending modulus for tubular cross-section
        return self.Sx

    @property
    def Rgyr(self):  # Radius of Gyration for circular tube
        return np.sqrt(self.Ixx / self.Area)

    @property
    def Klr(self):  # Klr buckling parameter
        return self.Kbuck * self.L / self.Rgyr


class IBeam(CrossSectionBase):
    def __init__(self, L_flange, t_flange, H_web, t_web, L=np.nan):
        self.Lf = L_flange
        self.tf = t_flange
        self.Hw = H_web
        self.tw = t_web
        self.H = H_web + 2 * t_flange
        self.L = L * np.ones(np.size(L_flange))

    @property
    def AreaFlange(self):  # Cross sectional area of tube
        return self.Lf * self.tf

    @property
    def AreaWeb(self):  # Cross sectional area of tube
        return self.Hw * self.tw

    @property
    def Area(self):  # Cross sectional area of tube
        return self.AreaWeb + 2 * self.AreaFlange

    @property
    def Ixx(self):  # 2nd area moment of inertia w.r.t. x-x axis running parallel to flange through CG
        return (self.Lf * self.H ** 3 - (self.Lf - self.tw) * self.Hw ** 3) / 12.0

    @property
    def Iyy(self):  # 2nd area moment of inertia w.r.t. z-z running through center of web
        return (2 * self.tw * self.Lf ** 3 + self.Hw * self.tw ** 3) / 12.0

    @property
    def J0(self):  # polar moment of inertia w.r.t. z-z axis (torsional)
        return 2 * self.Lf * self.tf ** 3 + self.H * self.tw ** 3

    @property
    def Asx(self):  # Shear Area for tubular cross-section
        return 1.64 * self.Lf * self.tf

    @property
    def Asy(self):  # Shear Area for tubular cross-section
        return self.tw * self.H

    @property
    def BdgMxx(self):  # Bending modulus for tubular cross-section
        return 2 * self.Ixx / self.H

    @property
    def BdgMyy(self):  # Bending modulus for tubular cross-section =BdgMxx
        return 2 * self.Iyy / self.Lf

    @property
    def TorsConst(self):  # Torsion shear constant for tubular cross-section
        return self.J0 / (1.28 * self.tf)

    @property
    def CG(self):
        return 0.5 * self.Hw + self.tf


class Square(CrossSectionBase):
    def __init__(self, a, t, L=np.nan):
        self.a = a
        self.t = t
        self.L = L * np.ones(np.size(a))

    @property
    def Area(self):  # Cross sectional area of tube
        return self.a ** 2 - (self.a - 2 * self.t) ** 2

    @property
    def Ixx(self):  # 2nd area moment of inertia w.r.t. x-x axis running parallel to flange through CG
        return (self.a ** 4 - (self.a - 2 * self.t) ** 4) / 12.0

    @property
    def Iyy(self):  # 2nd area moment of inertia w.r.t. z-z running through center of web
        return self.Ixx

    @property
    def J0(self):  # polar moment of inertia w.r.t. z-z axis (torsional)
        return self.t * (self.a - 2 * self.t) ** 3

    @property
    def Asx(self):  # Shear Area for tubular cross-section
        return self.Area / (2.39573 - 0.25009 * (self.t / self.a) - 7.89675 * (self.t / self.a) ** 2)

    @property
    def Asy(self):  # Shear Area for tubular cross-section
        return self.Asx

    @property
    def BdgMxx(self):  # Bending modulus for tubular cross-section
        return 2 * self.Ixx / self.a

    @property
    def BdgMyy(self):  # Bending modulus for tubular cross-section =BdgMxx
        return self.BdgMxx

    @property
    def TorsConst(self):  # Torsion shear constant for tubular cross-section
        return 2 * self.t * (self.a - self.t) ** 2

    @property
    def S(self):  # Bending modulus for tubular cross-section
        return self.Sx


class Rectangle(CrossSectionBase):
    def __init__(self, a, b, t, L=np.nan):
        self.a = a
        self.b = b
        self.t = t
        self.L = L * np.ones(np.size(a))

    @property
    def Area(self):  # Cross sectional area of tube
        return self.a * self.b - (self.a - 2 * self.t) * (self.b - 2 * self.t)

    @property
    def Ixx(self):  # 2nd area moment of inertia w.r.t. x-x axis running parallel to flange through CG
        return (self.a * self.b ** 3 - (self.a - 2 * self.t) * (self.b - 2 * self.t) * 3) / 12.0

    @property
    def Iyy(self):  # 2nd area moment of inertia w.r.t. z-z running through center of web
        return (self.b * self.a ** 3 - (self.b - 2 * self.t) * (self.a - 2 * self.t) * 3) / 12.0

    @property
    def J0(self):  # polar moment of inertia w.r.t. z-z axis (torsional)
        return 2 * self.t * (self.a - self.t) ** 2 * (self.b - self.t) ** 2 / (self.a + self.b - 2 * self.t)

    @property
    def Asx(self):  # Shear Area for tubular cross-section
        if self.a > self.b:
            return self.Area / (
                0.93498
                - 1.28084 * (self.t / self.b)
                + 1.36441 * (self.b / self.a)
                + 0.00295 * (self.a / self.b) ** 2
                + 0.25797 * (self.t * self.a / self.b ** 2)
            )
        else:
            return self.Area / (
                1.63544
                - 8.34935 * (self.t / self.a)
                + 0.60125 * (self.b / self.a)
                + 0.41403 * (self.b / self.a) ** 2
                + 4.95373 * (self.t * self.b / self.a ** 2)
            )

    @property
    def Asy(self):  # Shear Area for tubular cross-section
        if self.a > self.b:
            return self.Area / (
                1.63544
                - 8.34935 * (self.t / self.b)
                + 0.60125 * (self.a / self.b)
                + 0.41403 * (self.a / self.b) ** 2
                + 4.95373 * (self.t * self.a / self.b ** 2)
            )
        else:
            return self.Area / (
                0.93498
                - 1.28084 * (self.t / self.a)
                + 1.36441 * (self.a / self.b)
                + 0.00295 * (self.b / self.a) ** 2
                + 0.25797 * (self.t * self.b / self.a ** 2)
            )

    @property
    def BdgMxx(self):  # Bending modulus for tubular cross-section
        return 2 * self.Ixx / self.b

    @property
    def BdgMyy(self):  # Bending modulus for tubular cross-section =BdgMxx
        return 2 * self.Iyy / self.a

    @property
    def TorsConst(self):  # Torsion shear constant for tubular cross-section
        return 2 * self.t * (self.a - self.t) * (self.b - self.t)



"""Supply chain related infrastructure."""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2022, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"

from marmot import Agent, process


class SubstructureDelivery(Agent):
    """"""

    def __init__(self, component, num, deilvery_time, port, items, num_parallel=1):
        """
        Creates an instance of `SupplyChain`.

        Parameters
        ----------
        component : str
            Component to be delivered.
        num : int
            Number of component sets to be delivered.
        deilvery_time : int | float
            Time to between deliveries.
        num_parallel : int (optional)
            Number of parallel assembly lines.
            Default: 1
        """

        super().__init__(f"{component} Supply Chain")

        self.type = component
        self.num = num
        self.delivery_time = deilvery_time
        self.port = port
        self.items = items
        self.num_parallel = num_parallel

    @process
    def start(self):

        n = 0
        while n < self.num:
            yield self.task(
                f"Delivered {self.num_parallel} {self.type}(s)",
                self.delivery_time,
                cost=0,
            )

            for _ in range(self.num_parallel):
                for item in self.items:
                    self.port.put(item)

                n += 1



import unittest

import numpy as np
from openmdao.api import Problem
from wisdem.nrelcsm.nrel_csm_cost_2015 import Turbine_CostsSE_2015


class TestNewAssembly(unittest.TestCase):
    def setUp(self):
        turbine = Turbine_CostsSE_2015(verbosity=False)
        self.prob = Problem(turbine)
        self.prob.setup()

        self.prob["blade_mass"] = 17650.67  # inline with the windpact estimates
        self.prob["hub_mass"] = 31644.5
        self.prob["pitch_system_mass"] = 17004.0
        self.prob["spinner_mass"] = 1810.5
        self.prob["lss_mass"] = 31257.3
        # bearingsMass'] = 9731.41
        self.prob["main_bearing_mass"] = 9731.41 / 2
        self.prob["gearbox_mass"] = 30237.60
        self.prob["hss_mass"] = 1492.45
        self.prob["generator_mass"] = 16699.85
        self.prob["bedplate_mass"] = 93090.6
        self.prob["yaw_mass"] = 11878.24
        self.prob["tower_mass"] = 434559.0
        self.prob["converter_mass"] = 1000.0
        self.prob["hvac_mass"] = 1000.0
        self.prob["cover_mass"] = 1000.0
        self.prob["platforms_mass"] = 1000.0
        self.prob["transformer_mass"] = 1000.0

        # other inputs
        self.prob["machine_rating"] = 5000.0
        self.prob["blade_number"] = 3
        self.prob["crane"] = True
        self.prob["main_bearing_number"] = 2

    def test1(self):

        self.prob.run_model()

        self.assertEqual(np.round(self.prob["rotor_cost"], 2), 1292397.85)
        self.assertEqual(np.round(self.prob["generator_cost"], 2), 207078.14)
        self.assertEqual(np.round(self.prob["transformer_cost"], 2), 18800.00)
        self.assertEqual(np.round(self.prob["turbine_cost"], 2), 4404316.13)


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestNewAssembly))
    return suite


if __name__ == "__main__":
    result = unittest.TextTestRunner().run(suite())

    if result.wasSuccessful():
        exit(0)
    else:
        exit(1)



#!/usr/bin/env python
# encoding: utf-8
"""
airfoilprep.py

Created by Andrew Ning on 2012-04-16.
Copyright (c) NREL. All rights reserved.


Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

import copy

import numpy as np

# from scipy.interpolate import RectBivariateSpline


class Polar(object):
    """
    Defines section lift, drag, and pitching moment coefficients as a
    function of angle of attack at a particular Reynolds number.

    """

    def __init__(self, Re, alpha, cl, cd, cm):
        """Constructor

        Parameters
        ----------
        Re : float
            Reynolds number
        alpha : ndarray (deg)
            angle of attack
        cl : ndarray
            lift coefficient
        cd : ndarray
            drag coefficient
        cm : ndarray
            moment coefficient
        """

        self.Re = Re
        self.alpha = np.array(alpha)
        self.cl = np.array(cl)
        self.cd = np.array(cd)
        self.cm = np.array(cm)

    def blend(self, other, weight):
        """Blend this polar with another one with the specified weighting

        Parameters
        ----------
        other : Polar
            another Polar object to blend with
        weight : float
            blending parameter between 0 and 1.  0 returns self, whereas 1 returns other.

        Returns
        -------
        polar : Polar
            a blended Polar

        """

        # generate merged set of angles of attack - get unique values
        alpha = np.union1d(self.alpha, other.alpha)

        # truncate (TODO: could also have option to just use one of the polars for values out of range)
        min_alpha = max(self.alpha.min(), other.alpha.min())
        max_alpha = min(self.alpha.max(), other.alpha.max())
        alpha = alpha[np.logical_and(alpha >= min_alpha, alpha <= max_alpha)]
        # alpha = np.array([a for a in alpha if a >= min_alpha and a <= max_alpha])

        # interpolate to new alpha
        cl1 = np.interp(alpha, self.alpha, self.cl)
        cl2 = np.interp(alpha, other.alpha, other.cl)
        cd1 = np.interp(alpha, self.alpha, self.cd)
        cd2 = np.interp(alpha, other.alpha, other.cd)
        cm1 = np.interp(alpha, self.alpha, self.cm)
        cm2 = np.interp(alpha, other.alpha, other.cm)

        # linearly blend
        Re = self.Re + weight * (other.Re - self.Re)
        cl = cl1 + weight * (cl2 - cl1)
        cd = cd1 + weight * (cd2 - cd1)
        cm = cm1 + weight * (cm2 - cm1)

        return type(self)(Re, alpha, cl, cd, cm)

    def correction3D(self, r_over_R, chord_over_r, tsr, alpha_max_corr=30, alpha_linear_min=-5, alpha_linear_max=5):
        """Applies 3-D corrections for rotating sections from the 2-D data.

        Parameters
        ----------
        r_over_R : float
            local radial position / rotor radius
        chord_over_r : float
            local chord length / local radial location
        tsr : float
            tip-speed ratio
        alpha_max_corr : float, optional (deg)
            maximum angle of attack to apply full correction
        alpha_linear_min : float, optional (deg)
            angle of attack where linear portion of lift curve slope begins
        alpha_linear_max : float, optional (deg)
            angle of attack where linear portion of lift curve slope ends

        Returns
        -------
        polar : Polar
            A new Polar object corrected for 3-D effects

        Notes
        -----
        The Du-Selig method :cite:`Du1998A-3-D-stall-del` is used to correct lift, and
        the Eggers method :cite:`Eggers-Jr2003An-assessment-o` is used to correct drag.


        """

        # rename and convert units for convenience
        alpha = np.radians(self.alpha)
        cl_2d = self.cl
        cd_2d = self.cd
        alpha_max_corr = np.radians(alpha_max_corr)
        alpha_linear_min = np.radians(alpha_linear_min)
        alpha_linear_max = np.radians(alpha_linear_max)

        # parameters in Du-Selig model
        a = 1
        b = 1
        d = 1
        lam = tsr / (1 + tsr ** 2) ** 0.5  # modified tip speed ratio
        expon = d / lam / r_over_R
        expon_d = d / lam / r_over_R / 2.0

        # find linear region
        idx = np.logical_and(alpha >= alpha_linear_min, alpha <= alpha_linear_max)
        p = np.polyfit(alpha[idx], cl_2d[idx], 1)
        m = p[0]
        alpha0 = -p[1] / m

        # correction factor
        fcl = 1.0 / m * (1.6 * chord_over_r / 0.1267 * (a - chord_over_r ** expon) / (b + chord_over_r ** expon) - 1)
        fcd = (
            1.0 / m * (1.6 * chord_over_r / 0.1267 * (a - chord_over_r ** expon_d) / (b + chord_over_r ** expon_d) - 1)
        )

        # not sure where this adjustment comes from (besides AirfoilPrep spreadsheet of course)
        adj = ((np.pi / 2 - alpha) / (np.pi / 2 - alpha_max_corr)) ** 2
        adj[alpha <= alpha_max_corr] = 1.0

        # Du-Selig correction for lift
        cl_linear = m * (alpha - alpha0)
        cl_3d = cl_2d + fcl * (cl_linear - cl_2d) * adj

        # Du-Selig correction for drag
        cd0 = np.interp(0.0, alpha, cd_2d)
        dcd = cd_2d - cd0
        cd_3d = cd_2d + fcd * dcd

        # # Eggers 2003 correction for drag
        # delta_cl = cl_3d-cl_2d

        # delta_cd = delta_cl*(np.sin(alpha) - 0.12*np.cos(alpha))/(np.cos(alpha) + 0.12*np.sin(alpha))
        # cd_3d2 = cd_2d + delta_cd

        return type(self)(self.Re, np.degrees(alpha), cl_3d, cd_3d, self.cm)

    def extrapolate(self, cdmax, AR=None, cdmin=0.001, nalpha=15):
        """Extrapolates force coefficients up to +/- 180 degrees using Viterna's method
        :cite:`Viterna1982Theoretical-and`.

        Parameters
        ----------
        cdmax : float
            maximum drag coefficient
        AR : float, optional
            aspect ratio = (rotor radius / chord_75% radius)
            if provided, cdmax is computed from AR
        cdmin: float, optional
            minimum drag coefficient.  used to prevent negative values that can sometimes occur
            with this extrapolation method
        nalpha: int, optional
            number of points to add in each segment of Viterna method

        Returns
        -------
        polar : Polar
            a new Polar object

        Notes
        -----
        If the current polar already supplies data beyond 90 degrees then
        this method cannot be used in its current form and will just return itself.

        If AR is provided, then the maximum drag coefficient is estimated as

        >>> cdmax = 1.11 + 0.018*AR


        """

        if cdmin < 0:
            raise Exception("cdmin cannot be < 0")

        # lift coefficient adjustment to account for assymetry
        cl_adj = 0.7

        # estimate CD max
        if AR is not None:
            cdmax = 1.11 + 0.018 * AR
        self.cdmax = max(max(self.cd), cdmax)

        # extract matching info from ends
        alpha_high = np.radians(self.alpha[-1])
        cl_high = self.cl[-1]
        cd_high = self.cd[-1]
        cm_high = self.cm[-1]

        alpha_low = np.radians(self.alpha[0])
        cl_low = self.cl[0]
        cd_low = self.cd[0]

        if alpha_high > np.pi / 2:
            raise Exception("alpha[-1] > pi/2")
            return self
        if alpha_low < -np.pi / 2:
            raise Exception("alpha[0] < -pi/2")
            return self

        # parameters used in model
        sa = np.sin(alpha_high)
        ca = np.cos(alpha_high)
        self.A = (cl_high - self.cdmax * sa * ca) * sa / ca ** 2
        self.B = (cd_high - self.cdmax * sa * sa) / ca

        # alpha_high <-> 90
        alpha1 = np.linspace(alpha_high, np.pi / 2, nalpha)
        alpha1 = alpha1[1:]  # remove first element so as not to duplicate when concatenating
        cl1, cd1 = self.__Viterna(alpha1, 1.0)

        # 90 <-> 180-alpha_high
        alpha2 = np.linspace(np.pi / 2, np.pi - alpha_high, nalpha)
        alpha2 = alpha2[1:]
        cl2, cd2 = self.__Viterna(np.pi - alpha2, -cl_adj)

        # 180-alpha_high <-> 180
        alpha3 = np.linspace(np.pi - alpha_high, np.pi, nalpha)
        alpha3 = alpha3[1:]
        cl3, cd3 = self.__Viterna(np.pi - alpha3, 1.0)
        cl3 = (alpha3 - np.pi) / alpha_high * cl_high * cl_adj  # override with linear variation

        if alpha_low <= -alpha_high:
            alpha4 = []
            cl4 = []
            cd4 = []
            alpha5max = alpha_low
        else:
            # -alpha_high <-> alpha_low
            # Note: this is done slightly differently than AirfoilPrep for better continuity
            alpha4 = np.linspace(-alpha_high, alpha_low, nalpha)
            alpha4 = alpha4[1:-2]  # also remove last element for concatenation for this case
            cl4 = -cl_high * cl_adj + (alpha4 + alpha_high) / (alpha_low + alpha_high) * (cl_low + cl_high * cl_adj)
            cd4 = cd_low + (alpha4 - alpha_low) / (-alpha_high - alpha_low) * (cd_high - cd_low)
            alpha5max = -alpha_high

        # -90 <-> -alpha_high
        alpha5 = np.linspace(-np.pi / 2, alpha5max, nalpha)
        alpha5 = alpha5[1:]
        cl5, cd5 = self.__Viterna(-alpha5, -cl_adj)

        # -180+alpha_high <-> -90
        alpha6 = np.linspace(-np.pi + alpha_high, -np.pi / 2, nalpha)
        alpha6 = alpha6[1:]
        cl6, cd6 = self.__Viterna(alpha6 + np.pi, cl_adj)

        # -180 <-> -180 + alpha_high
        alpha7 = np.linspace(-np.pi, -np.pi + alpha_high, nalpha)
        cl7, cd7 = self.__Viterna(alpha7 + np.pi, 1.0)
        cl7 = (alpha7 + np.pi) / alpha_high * cl_high * cl_adj  # linear variation

        alpha = np.concatenate((alpha7, alpha6, alpha5, alpha4, np.radians(self.alpha), alpha1, alpha2, alpha3))
        cl = np.concatenate((cl7, cl6, cl5, cl4, self.cl, cl1, cl2, cl3))
        cd = np.concatenate((cd7, cd6, cd5, cd4, self.cd, cd1, cd2, cd3))

        cd = np.maximum(cd, cdmin)  # don't allow negative drag coefficients

        # Setup alpha and cm to be used in extrapolation
        cm1_alpha = np.floor(self.alpha[0] / 10.0) * 10.0
        cm2_alpha = np.ceil(self.alpha[-1] / 10.0) * 10.0
        alpha_num = abs(int((-180.0 - cm1_alpha) / 10.0 - 1))
        alpha_cm1 = np.linspace(-180.0, cm1_alpha, alpha_num)
        alpha_cm2 = np.linspace(cm2_alpha, 180.0, int((180.0 - cm2_alpha) / 10.0 + 1))
        alpha_cm = np.concatenate(
            (alpha_cm1, self.alpha, alpha_cm2)
        )  # Specific alpha values are needed for cm function to work
        cm1 = np.zeros(len(alpha_cm1))
        cm2 = np.zeros(len(alpha_cm2))
        cm_ext = np.concatenate((cm1, self.cm, cm2))
        if np.count_nonzero(self.cm) > 0:
            cmCoef = self.__CMCoeff(cl_high, cd_high, cm_high)  # get cm coefficient
            cl_cm = np.interp(alpha_cm, np.degrees(alpha), cl)  # get cl for applicable alphas
            cd_cm = np.interp(alpha_cm, np.degrees(alpha), cd)  # get cd for applicable alphas
            alpha_low_deg = self.alpha[0]
            alpha_high_deg = self.alpha[-1]
            for i in range(len(alpha_cm)):
                cm_new = self.__getCM(i, cmCoef, alpha_cm, cl_cm, cd_cm, alpha_low_deg, alpha_high_deg)
                if cm_new is None:
                    pass  # For when it reaches the range of cm's that the user provides
                else:
                    cm_ext[i] = cm_new
        cm = np.interp(np.degrees(alpha), alpha_cm, cm_ext)
        return type(self)(self.Re, np.degrees(alpha), cl, cd, cm)

    def __Viterna(self, alpha, cl_adj):
        """private method to perform Viterna extrapolation"""

        alpha = np.maximum(alpha, 0.0001)  # prevent divide by zero

        cl = self.cdmax / 2 * np.sin(2 * alpha) + self.A * np.cos(alpha) ** 2 / np.sin(alpha)
        cl = cl * cl_adj

        cd = self.cdmax * np.sin(alpha) ** 2 + self.B * np.cos(alpha)

        return cl, cd

    def __CMCoeff(self, cl_high, cd_high, cm_high):
        """private method to obtain CM0 and CMCoeff"""

        found_zero_lift = False

        for i in range(len(self.cm) - 1):
            if abs(self.alpha[i]) < 20.0 and self.cl[i] <= 0 and self.cl[i + 1] >= 0:
                p = -self.cl[i] / (self.cl[i + 1] - self.cl[i])
                cm0 = self.cm[i] + p * (self.cm[i + 1] - self.cm[i])
                found_zero_lift = True
                break

        if not found_zero_lift:
            p = -self.cl[0] / (self.cl[1] - self.cl[0])
            cm0 = self.cm[0] + p * (self.cm[1] - self.cm[0])
        self.cm0 = cm0
        alpha_high = np.radians(self.alpha[-1])
        XM = (-cm_high + cm0) / (cl_high * np.cos(alpha_high) + cd_high * np.sin(alpha_high))
        cmCoef = (XM - 0.25) / np.tan((alpha_high - np.pi / 2))
        return cmCoef

    def __getCM(self, i, cmCoef, alpha, cl_ext, cd_ext, alpha_low_deg, alpha_high_deg):
        """private method to extrapolate Cm"""

        cm_new = 0
        if alpha[i] >= alpha_low_deg and alpha[i] <= alpha_high_deg:
            return
        if alpha[i] > -165 and alpha[i] < 165:
            if abs(alpha[i]) < 0.01:
                cm_new = self.cm0
            else:
                if alpha[i] > 0:
                    x = cmCoef * np.tan(np.radians(alpha[i]) - np.pi / 2) + 0.25
                    cm_new = self.cm0 - x * (
                        cl_ext[i] * np.cos(np.radians(alpha[i])) + cd_ext[i] * np.sin(np.radians(alpha[i]))
                    )
                else:
                    x = cmCoef * np.tan(-np.radians(alpha[i]) - np.pi / 2) + 0.25
                    cm_new = -(
                        self.cm0
                        - x * (-cl_ext[i] * np.cos(-np.radians(alpha[i])) + cd_ext[i] * np.sin(-np.radians(alpha[i])))
                    )
        else:
            if alpha[i] == 165:
                cm_new = -0.4
            elif alpha[i] == 170:
                cm_new = -0.5
            elif alpha[i] == 175:
                cm_new = -0.25
            elif alpha[i] == 180:
                cm_new = 0
            elif alpha[i] == -165:
                cm_new = 0.35
            elif alpha[i] == -170:
                cm_new = 0.4
            elif alpha[i] == -175:
                cm_new = 0.2
            elif alpha[i] == -180:
                cm_new = 0
            else:
                print("Angle encountered for which there is no CM table value " "(near +/-180 deg). Program will stop.")
        return cm_new

    def unsteadyparam(self, alpha_linear_min=-5, alpha_linear_max=5):
        """compute unsteady aero parameters used in AeroDyn input file

        Parameters
        ----------
        alpha_linear_min : float, optional (deg)
            angle of attack where linear portion of lift curve slope begins
        alpha_linear_max : float, optional (deg)
            angle of attack where linear portion of lift curve slope ends

        Returns
        -------
        aerodynParam : tuple of floats
            (control setting, stall angle, alpha for 0 cn, cn slope,
            cn at stall+, cn at stall-, alpha for min CD, min(CD))

        """

        alpha = np.radians(self.alpha)
        cl = self.cl
        cd = self.cd

        alpha_linear_min = np.radians(alpha_linear_min)
        alpha_linear_max = np.radians(alpha_linear_max)

        cn = cl * np.cos(alpha) + cd * np.sin(alpha)

        # find linear region
        idx = np.logical_and(alpha >= alpha_linear_min, alpha <= alpha_linear_max)

        # checks for inppropriate data (like cylinders)
        if len(idx) < 10 or len(np.unique(cl)) < 10:
            return (
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
            )

        # linear fit
        p = np.polyfit(alpha[idx], cn[idx], 1)
        m = p[0]
        alpha0 = -p[1] / m

        # find cn at stall locations
        alphaUpper = np.radians(np.arange(40.0))
        alphaLower = np.radians(np.arange(5.0, -40.0, -1))
        cnUpper = np.interp(alphaUpper, alpha, cn)
        cnLower = np.interp(alphaLower, alpha, cn)
        cnLinearUpper = m * (alphaUpper - alpha0)
        cnLinearLower = m * (alphaLower - alpha0)
        deviation = 0.05  # threshold for cl in detecting stall

        alphaU = np.interp(deviation, cnLinearUpper - cnUpper, alphaUpper)
        alphaL = np.interp(deviation, cnLower - cnLinearLower, alphaLower)

        # compute cn at stall according to linear fit
        cnStallUpper = m * (alphaU - alpha0)
        cnStallLower = m * (alphaL - alpha0)

        # find min cd
        minIdx = cd.argmin()

        # return: control setting, stall angle, alpha for 0 cn, cn slope,
        #         cn at stall+, cn at stall-, alpha for min CD, min(CD)
        return (0.0, np.degrees(alphaU), np.degrees(alpha0), m, cnStallUpper, cnStallLower, alpha[minIdx], cd[minIdx])

    def plot(self):
        """plot cl/cd/cm polar

        Returns
        -------
        figs : list of figure handles

        """
        import matplotlib.pyplot as plt

        p = self

        figs = []

        # plot cl
        fig = plt.figure()
        figs.append(fig)
        ax = fig.add_subplot(111)
        plt.plot(p.alpha, p.cl, label="Re = " + str(p.Re / 1e6) + " million")
        ax.set_xlabel("angle of attack (deg)")
        ax.set_ylabel("lift coefficient")
        ax.legend(loc="best")

        # plot cd
        fig = plt.figure()
        figs.append(fig)
        ax = fig.add_subplot(111)
        ax.plot(p.alpha, p.cd, label="Re = " + str(p.Re / 1e6) + " million")
        ax.set_xlabel("angle of attack (deg)")
        ax.set_ylabel("drag coefficient")
        ax.legend(loc="best")

        # plot cm
        fig = plt.figure()
        figs.append(fig)
        ax = fig.add_subplot(111)
        ax.plot(p.alpha, p.cm, label="Re = " + str(p.Re / 1e6) + " million")
        ax.set_xlabel("angle of attack (deg)")
        ax.set_ylabel("moment coefficient")
        ax.legend(loc="best")

        return figs


class Airfoil(object):
    """A collection of Polar objects at different Reynolds numbers"""

    def __init__(self, polars):
        """Constructor

        Parameters
        ----------
        polars : list(Polar)
            list of Polar objects

        """

        # sort by Reynolds number
        self.polars = sorted(polars, key=lambda p: p.Re)

        # save type of polar we are using
        self.polar_type = polars[0].__class__

    @classmethod
    def initFromAerodynFile(cls, aerodynFile, polarType=Polar):
        """Construct Airfoil object from AeroDyn file

        Parameters
        ----------
        aerodynFile : str
            path/name of a properly formatted Aerodyn file

        Returns
        -------
        obj : Airfoil

        """
        # initialize
        polars = []

        # open aerodyn file
        f = open(aerodynFile, "r")

        # skip through header
        f.readline()
        description = f.readline().rstrip()  # remove newline
        f.readline()
        numTables = int(f.readline().split()[0])

        # loop through tables
        for i in range(numTables):

            # read Reynolds number
            Re = float(f.readline().split()[0]) * 1e6

            # read Aerodyn parameters
            param = [0] * 8
            for j in range(8):
                param[j] = float(f.readline().split()[0])

            alpha = []
            cl = []
            cd = []
            cm = []

            # read polar information line by line
            while True:
                line = f.readline()
                if "EOT" in line:
                    break
                data = [float(s) for s in line.split()]
                if len(data) < 4:
                    raise ValueError(f"Error: Expected 4 columns of data but found, {data}")

                alpha.append(data[0])
                cl.append(data[1])
                cd.append(data[2])
                cm.append(data[3])

            polars.append(polarType(Re, alpha, cl, cd, cm))

        f.close()

        return cls(polars)

    def getPolar(self, Re):
        """Gets a Polar object for this airfoil at the specified Reynolds number.

        Parameters
        ----------
        Re : float
            Reynolds number

        Returns
        -------
        obj : Polar
            a Polar object

        Notes
        -----
        Interpolates as necessary. If Reynolds number is larger than or smaller than
        the stored Polars, it returns the Polar with the closest Reynolds number.

        """

        p = self.polars

        if Re <= p[0].Re:
            return copy.deepcopy(p[0])

        elif Re >= p[-1].Re:
            return copy.deepcopy(p[-1])

        else:
            Relist = [pp.Re for pp in p]
            i = np.searchsorted(Relist, Re)
            weight = (Re - Relist[i - 1]) / (Relist[i] - Relist[i - 1])
            return p[i - 1].blend(p[i], weight)

    def blend(self, other, weight):
        """Blend this Airfoil with another one with the specified weighting.


        Parameters
        ----------
        other : Airfoil
            other airfoil to blend with
        weight : float
            blending parameter between 0 and 1.  0 returns self, whereas 1 returns other.

        Returns
        -------
        obj : Airfoil
            a blended Airfoil object

        Notes
        -----
        First finds the unique Reynolds numbers.  Evaluates both sets of polars
        at each of the Reynolds numbers, then blends at each Reynolds number.

        """

        # combine Reynolds numbers
        Relist1 = [p.Re for p in self.polars]
        Relist2 = [p.Re for p in other.polars]
        Relist = np.union1d(Relist1, Relist2)

        # blend polars
        n = len(Relist)
        polars = [0] * n
        for i in range(n):
            p1 = self.getPolar(Relist[i])
            p2 = other.getPolar(Relist[i])
            polars[i] = p1.blend(p2, weight)

        return Airfoil(polars)

    def correction3D(self, r_over_R, chord_over_r, tsr, alpha_max_corr=30, alpha_linear_min=-5, alpha_linear_max=5):
        """apply 3-D rotational corrections to each polar in airfoil

        Parameters
        ----------
        r_over_R : float
            radial position / rotor radius
        chord_over_r : float
            local chord / local radius
        tsr : float
            tip-speed ratio
        alpha_max_corr : float, optional (deg)
            maximum angle of attack to apply full correction
        alpha_linear_min : float, optional (deg)
            angle of attack where linear portion of lift curve slope begins
        alpha_linear_max : float, optional (deg)
            angle of attack where linear portion of lift curve slope ends

        Returns
        -------
        airfoil : Airfoil
            airfoil with 3-D corrections

        See Also
        --------
        Polar.correction3D : apply 3-D corrections for a Polar

        """

        n = len(self.polars)
        polars = [0] * n
        for idx, p in enumerate(self.polars):
            polars[idx] = p.correction3D(
                r_over_R, chord_over_r, tsr, alpha_max_corr, alpha_linear_min, alpha_linear_max
            )

        return Airfoil(polars)

    def extrapolate(self, cdmax, AR=None, cdmin=0.001):
        """apply high alpha extensions to each polar in airfoil

        Parameters
        ----------
        cdmax : float
            maximum drag coefficient
        AR : float, optional
            blade aspect ratio (rotor radius / chord at 75% radius).  if included
            it is used to estimate cdmax
        cdmin: minimum drag coefficient

        Returns
        -------
        airfoil : Airfoil
            airfoil with +/-180 degree extensions

        See Also
        --------
        Polar.extrapolate : extrapolate a Polar to high angles of attack

        """

        n = len(self.polars)
        polars = [0] * n
        for idx, p in enumerate(self.polars):
            polars[idx] = p.extrapolate(cdmax, AR, cdmin)

        return Airfoil(polars)

    def interpToCommonAlpha(self, alpha=None):
        """Interpolates all polars to a common set of angles of attack

        Parameters
        ----------
        alpha : ndarray, optional
            common set of angles of attack to use.  If None a union of
            all angles of attack in the polars is used.

        """

        if alpha is None:
            # union of angle of attacks
            alpha = []
            for p in self.polars:
                alpha = np.union1d(alpha, p.alpha)

        # interpolate each polar to new alpha
        n = len(self.polars)
        polars = [0] * n
        for idx, p in enumerate(self.polars):
            cl = np.interp(alpha, p.alpha, p.cl)
            cd = np.interp(alpha, p.alpha, p.cd)
            cm = np.interp(alpha, p.alpha, p.cm)
            polars[idx] = self.polar_type(p.Re, alpha, cl, cd, cm)

        return Airfoil(polars)

    def writeToAerodynFile(self, filename):
        """Write the airfoil section data to a file using AeroDyn input file style.

        Parameters
        ----------
        filename : str
            name (+ relative path) of where to write file

        """

        # aerodyn and wtperf require common set of angles of attack
        af = self.interpToCommonAlpha()

        f = open(filename, "w")

        f.write("AeroDyn airfoil file.")
        f.write("Compatible with AeroDyn v13.0.")
        f.write("Generated by airfoilprep.py")
        f.write("{0:<10d}\t\t{1:40}".format(len(af.polars), "Number of airfoil tables in this file"))
        for p in af.polars:
            f.write("{0:<10f}\t{1:40}".format(p.Re / 1e6, "Reynolds number in millions."))
            param = p.unsteadyparam()
            f.write("{0:<10f}\t{1:40}".format(param[0], "Control setting"))
            f.write("{0:<10f}\t{1:40}".format(param[1], "Stall angle (deg)"))
            f.write("{0:<10f}\t{1:40}".format(param[2], "Angle of attack for zero Cn for linear Cn curve (deg)"))
            f.write("{0:<10f}\t{1:40}".format(param[3], "Cn slope for zero lift for linear Cn curve (1/rad)"))
            f.write(
                "{0:<10f}\t{1:40}".format(
                    param[4], "Cn at stall value for positive angle of attack for linear Cn curve"
                )
            )
            f.write(
                "{0:<10f}\t{1:40}".format(
                    param[5], "Cn at stall value for negative angle of attack for linear Cn curve"
                )
            )
            f.write("{0:<10f}\t{1:40}".format(param[6], "Angle of attack for minimum CD (deg)"))
            f.write("{0:<10f}\t{1:40}".format(param[7], "Minimum CD value"))
            for a, cl, cd, cm in zip(p.alpha, p.cl, p.cd, p.cm):
                f.write("{:<10f}\t{:<10f}\t{:<10f}\t{:<10f}".format(a, cl, cd, cm))
            f.write("EOT")
        f.close()

    def createDataGrid(self):
        """interpolate airfoil data onto uniform alpha-Re grid.

        Returns
        -------
        alpha : ndarray (deg)
            a common set of angles of attack (union of all polars)
        Re : ndarray
            all Reynolds numbers defined in the polars
        cl : ndarray
            lift coefficient 2-D array with shape (alpha.size, Re.size)
            cl[i, j] is the lift coefficient at alpha[i] and Re[j]
        cd : ndarray
            drag coefficient 2-D array with shape (alpha.size, Re.size)
            cd[i, j] is the drag coefficient at alpha[i] and Re[j]

        """

        af = self.interpToCommonAlpha()
        polarList = af.polars

        # angle of attack is already same for each polar
        alpha = polarList[0].alpha

        # all Reynolds numbers
        Re = [p.Re for p in polarList]

        # fill in cl, cd grid
        cl = np.zeros((len(alpha), len(Re)))
        cd = np.zeros((len(alpha), len(Re)))
        cm = np.zeros((len(alpha), len(Re)))

        for (idx, p) in enumerate(polarList):
            cl[:, idx] = p.cl
            cd[:, idx] = p.cd
            cm[:, idx] = p.cm

        return alpha, Re, cl, cd, cm

    def plot(self, single_figure=True):
        """plot cl/cd/cm polars

        Parameters
        ----------
        single_figure : bool
            True  : plot all cl on the same figure (same for cd,cm)
            False : plot all cl/cd/cm on separate figures

        Returns
        -------
        figs : list of figure handles

        """

        import matplotlib.pyplot as plt

        figs = []

        # if in single figure mode (default)
        if single_figure:
            # generate figure handles
            fig1 = plt.figure()
            ax1 = fig1.add_subplot(111)
            figs.append(fig1)

            fig2 = plt.figure()
            ax2 = fig2.add_subplot(111)
            figs.append(fig2)

            # loop through all polars to see if we need to generate handles for cm figs
            for p in self.polars:
                if p.useCM == True:
                    fig3 = plt.figure()
                    ax3 = fig3.add_subplot(111)
                    figs.append(fig3)
                    break

            # loop through polars and plot
            for p in self.polars:
                # plot cl
                ax1.plot(p.alpha, p.cl, label="Re = " + str(p.Re / 1e6) + " million")
                ax1.set_xlabel("angle of attack (deg)")
                ax1.set_ylabel("lift coefficient")
                ax1.legend(loc="best")

                # plot cd
                ax2.plot(p.alpha, p.cd, label="Re = " + str(p.Re / 1e6) + " million")
                ax2.set_xlabel("angle of attack (deg)")
                ax2.set_ylabel("drag coefficient")
                ax2.legend(loc="best")

                # plot cm
                ax3.plot(p.alpha, p.cm, label="Re = " + str(p.Re / 1e6) + " million")
                ax3.set_xlabel("angle of attack (deg)")
                ax3.set_ylabel("moment coefficient")
                ax3.legend(loc="best")

        # otherwise, multi figure mode -- plot all on separate figures
        else:
            for p in self.polars:
                fig = plt.figure()
                figs.append(fig)
                ax = fig.add_subplot(111)
                ax.plot(p.alpha, p.cl, label="Re = " + str(p.Re / 1e6) + " million")
                ax.set_xlabel("angle of attack (deg)")
                ax.set_ylabel("lift coefficient")
                ax.legend(loc="best")

                fig = plt.figure()
                figs.append(fig)
                ax = fig.add_subplot(111)
                ax.plot(p.alpha, p.cd, label="Re = " + str(p.Re / 1e6) + " million")
                ax.set_xlabel("angle of attack (deg)")
                ax.set_ylabel("drag coefficient")
                ax.legend(loc="best")

                fig = plt.figure()
                figs.append(fig)
                ax = fig.add_subplot(111)
                ax.plot(p.alpha, p.cm, label="Re = " + str(p.Re / 1e6) + " million")
                ax.set_xlabel("angle of attack (deg)")
                ax.set_ylabel("moment coefficient")
                ax.legend(loc="best")

        return figs

    # def evaluate(self, alpha, Re):
    #     """Get lift/drag coefficient at the specified angle of attack and Reynolds number

    #     Parameters
    #     ----------
    #     alpha : float (rad)
    #         angle of attack (in Radians!)
    #     Re : float
    #         Reynolds number

    #     Returns
    #     -------
    #     cl : float
    #         lift coefficient
    #     cd : float
    #         drag coefficient

    #     Notes
    #     -----
    #     Uses a spline so that output is continuously differentiable
    #     also uses a small amount of smoothing to help remove spurious multiple solutions

    #     """

    #     # setup spline if necessary
    #     if self.need_to_setup_spline:
    #         alpha_v, Re_v, cl_M, cd_M = self.createDataGrid()
    #         alpha_v = np.radians(alpha_v)

    #         # special case if zero or one Reynolds number (need at least two for bivariate spline)
    #         if len(Re_v) < 2:
    #             Re_v = [1e1, 1e15]
    #             cl_M = np.c_[cl_M, cl_M]
    #             cd_M = np.c_[cd_M, cd_M]

    #         kx = min(len(alpha_v)-1, 3)
    #         ky = min(len(Re_v)-1, 3)

    #         self.cl_spline = RectBivariateSpline(alpha_v, Re_v, cl_M, kx=kx, ky=ky, s=0.1)
    #         self.cd_spline = RectBivariateSpline(alpha_v, Re_v, cd_M, kx=kx, ky=ky, s=0.001)
    #         self.need_to_setup_spline = False

    #     # evaluate spline --- index to make scalar

    #     cl = self.cl_spline.ev(alpha, Re)[0]
    #     cd = self.cd_spline.ev(alpha, Re)[0]

    #     return cl, cd


if __name__ == "__main__":

    import os
    from argparse import ArgumentParser, RawTextHelpFormatter

    # setup command line arguments
    parser = ArgumentParser(
        formatter_class=RawTextHelpFormatter, description="Preprocessing airfoil data for wind turbine applications."
    )
    parser.add_argument("src_file", type=str, help="source file")
    parser.add_argument(
        "--stall3D", type=str, nargs=3, metavar=("r/R", "c/r", "tsr"), help="2D data -> apply 3D corrections"
    )
    parser.add_argument("--extrap", type=str, nargs=1, metavar=("cdmax"), help="3D data -> high alpha extrapolations")
    parser.add_argument(
        "--blend",
        type=str,
        nargs=2,
        metavar=("otherfile", "weight"),
        help="blend 2 files weight 0: sourcefile, weight 1: otherfile",
    )
    parser.add_argument("--out", type=str, help="output file")
    parser.add_argument("--plot", action="store_true", help="plot data using matplotlib")
    parser.add_argument(
        "--common",
        action="store_true",
        help="interpolate the data at different Reynolds numbers to a common set of angles of attack",
    )

    # parse command line arguments
    args = parser.parse_args()
    fileOut = args.out

    if args.plot:
        import matplotlib.pyplot as plt

    # perform actions
    if args.stall3D is not None:

        if fileOut is None:
            name, ext = os.path.splitext(args.src_file)
            fileOut = name + "_3D" + ext

        af = Airfoil.initFromAerodynFile(args.src_file)
        floats = [float(var) for var in args.stall3D]
        af3D = af.correction3D(*floats)

        if args.common:
            af3D = af3D.interpToCommonAlpha()

        af3D.writeToAerodynFile(fileOut)

        if args.plot:

            for p, p3D in zip(af.polars, af3D.polars):
                # plt.figure(figsize=(6.0, 2.6))
                # plt.subplot(121)
                plt.figure()
                plt.plot(p.alpha, p.cl, "k", label="2D")
                plt.plot(p3D.alpha, p3D.cl, "r", label="3D")
                plt.xlabel("angle of attack (deg)")
                plt.ylabel("lift coefficient")
                plt.legend(loc="lower right")

                # plt.subplot(122)
                plt.figure()
                plt.plot(p.alpha, p.cd, "k", label="2D")
                plt.plot(p3D.alpha, p3D.cd, "r", label="3D")
                plt.xlabel("angle of attack (deg)")
                plt.ylabel("drag coefficient")
                plt.legend(loc="upper center")

                # plt.tight_layout()
                # plt.savefig('/Users/sning/Dropbox/NREL/SysEng/airfoilpreppy/docs/images/stall3d.pdf')

            plt.show()

    elif args.extrap is not None:

        if fileOut is None:
            name, ext = os.path.splitext(args.src_file)
            fileOut = name + "_extrap" + ext

        af = Airfoil.initFromAerodynFile(args.src_file)

        afext = af.extrapolate(float(args.extrap[0]))

        if args.common:
            afext = afext.interpToCommonAlpha()

        afext.writeToAerodynFile(fileOut)

        if args.plot:

            for p, pext in zip(af.polars, afext.polars):
                # plt.figure(figsize=(6.0, 2.6))
                # plt.subplot(121)
                plt.figure()
                (p1,) = plt.plot(pext.alpha, pext.cl, "r")
                (p2,) = plt.plot(p.alpha, p.cl, "k")
                plt.xlabel("angle of attack (deg)")
                plt.ylabel("lift coefficient")
                plt.legend([p2, p1], ["orig", "extrap"], loc="upper right")

                # plt.subplot(122)
                plt.figure()
                (p1,) = plt.plot(pext.alpha, pext.cd, "r")
                (p2,) = plt.plot(p.alpha, p.cd, "k")
                plt.xlabel("angle of attack (deg)")
                plt.ylabel("drag coefficient")
                plt.legend([p2, p1], ["orig", "extrap"], loc="lower right")

                plt.figure()
                (p1,) = plt.plot(pext.alpha, pext.cm, "r")
                (p2,) = plt.plot(p.alpha, p.cm, "k")
                plt.xlabel("angle of attack (deg)")
                plt.ylabel("moment coefficient")
                plt.legend([p2, p1], ["orig", "extrap"], loc="upper right")

                # plt.tight_layout()
                # plt.savefig('/Users/sning/Dropbox/NREL/SysEng/airfoilpreppy/docs/images/extrap.pdf')

            plt.show()

    elif args.blend is not None:

        if fileOut is None:
            name1, ext = os.path.splitext(args.src_file)
            name2, ext = os.path.splitext(os.path.basename(args.blend[0]))
            fileOut = name1 + "+" + name2 + "_blend" + args.blend[1] + ext

        af1 = Airfoil.initFromAerodynFile(args.src_file)
        af2 = Airfoil.initFromAerodynFile(args.blend[0])
        afOut = af1.blend(af2, float(args.blend[1]))

        if args.common:
            afOut = afOut.interpToCommonAlpha()

        afOut.writeToAerodynFile(fileOut)

        if args.plot:

            for p in afOut.polars:
                fig = plt.figure()
                ax = fig.add_subplot(111)
                plt.plot(p.alpha, p.cl, "k")
                plt.xlabel("angle of attack (deg)")
                plt.ylabel("lift coefficient")
                plt.text(0.6, 0.2, "Re = " + str(p.Re / 1e6) + " million", transform=ax.transAxes)

                fig = plt.figure()
                ax = fig.add_subplot(111)
                plt.plot(p.alpha, p.cd, "k")
                plt.xlabel("angle of attack (deg)")
                plt.ylabel("drag coefficient")
                plt.text(0.2, 0.8, "Re = " + str(p.Re / 1e6) + " million", transform=ax.transAxes)

                fig = plt.figure()
                ax = fig.add_subplot(111)
                plt.plot(p.alpha, p.cm, "k")
                plt.xlabel("angle of attack (deg)")
                plt.ylabel("moment coefficient")
                plt.text(0.2, 0.8, "Re = " + str(p.Re / 1e6) + " million", transform=ax.transAxes)

            plt.show()



import os

import numpy as np
import openmdao.api as om
from scipy.interpolate import PchipInterpolator


class ParametrizeBladeAero(om.ExplicitComponent):
    # Openmdao component to parameterize distributed quantities for the outer shape of the wind turbine rotor blades
    def initialize(self):
        self.options.declare("rotorse_options")
        self.options.declare("opt_options")

    def setup(self):
        rotorse_options = self.options["rotorse_options"]
        self.opt_options = self.options["opt_options"]
        n_span = rotorse_options["n_span"]
        self.n_opt_twist = n_opt_twist = self.opt_options["design_variables"]["blade"]["aero_shape"]["twist"]["n_opt"]
        self.n_opt_chord = n_opt_chord = self.opt_options["design_variables"]["blade"]["aero_shape"]["chord"]["n_opt"]

        # Inputs
        self.add_input(
            "s",
            val=np.zeros(n_span),
            desc="1D array of the non-dimensional spanwise grid defined along blade axis (0-blade root, 1-blade tip)",
        )
        # # Blade twist
        # self.add_input(
        #     "twist_original",
        #     val=np.zeros(n_span),
        #     units="rad",
        #     desc="1D array of the twist values defined along blade span. The twist is the one defined in the yaml.",
        # )
        self.add_input(
            "s_opt_twist",
            val=np.zeros(n_opt_twist),
            desc="1D array of the non-dimensional spanwise grid defined along blade axis to optimize the blade twist angle",
        )
        self.add_input(
            "twist_opt",
            val=np.ones(n_opt_twist),
            units="rad",
            desc="1D array of the twist angle being optimized at the n_opt locations.",
        )
        # Blade chord
        # self.add_input(
        #     "chord_original",
        #     val=np.zeros(n_span),
        #     units="m",
        #     desc="1D array of the chord values defined along blade span. The chord is the one defined in the yaml.",
        # )
        self.add_input(
            "s_opt_chord",
            val=np.zeros(n_opt_chord),
            desc="1D array of the non-dimensional spanwise grid defined along blade axis to optimize the blade chord",
        )
        self.add_input(
            "chord_opt",
            val=np.ones(n_opt_chord),
            units="m",
            desc="1D array of the chord being optimized at the n_opt locations.",
        )

        # Outputs
        self.add_output(
            "twist_param",
            val=np.zeros(n_span),
            units="rad",
            desc="1D array of the twist values defined along blade span. The twist is the result of the parameterization.",
        )
        self.add_output(
            "chord_param",
            val=np.zeros(n_span),
            units="m",
            desc="1D array of the chord values defined along blade span. The chord is the result of the parameterization.",
        )
        self.add_output(
            "max_chord_constr",
            val=np.zeros(n_opt_chord),
            desc="1D array of the ratio between chord values and maximum chord along blade span.",
        )

    def compute(self, inputs, outputs):

        spline = PchipInterpolator
        twist_spline = spline(inputs["s_opt_twist"], inputs["twist_opt"])
        outputs["twist_param"] = twist_spline(inputs["s"])
        chord_spline = spline(inputs["s_opt_chord"], inputs["chord_opt"])
        outputs["chord_param"] = chord_spline(inputs["s"])
        chord_opt = spline(inputs["s"], outputs["chord_param"])
        max_chord = self.opt_options["constraints"]["blade"]["chord"]["max"]
        outputs["max_chord_constr"] = chord_opt(inputs["s_opt_chord"]) / max_chord


class ParametrizeBladeStruct(om.ExplicitComponent):
    # Openmdao component to parameterize distributed quantities for the structural design of the wind turbine rotor blades
    def initialize(self):
        self.options.declare("rotorse_options")
        self.options.declare("opt_options")

    def setup(self):
        rotorse_options = self.options["rotorse_options"]
        self.opt_options = opt_options = self.options["opt_options"]
        self.n_span = n_span = rotorse_options["n_span"]
        self.n_layers = n_layers = rotorse_options["n_layers"]
        self.n_opt_spar_cap_ss = n_opt_spar_cap_ss = opt_options["design_variables"]["blade"]["structure"][
            "spar_cap_ss"
        ]["n_opt"]
        self.n_opt_spar_cap_ps = n_opt_spar_cap_ps = opt_options["design_variables"]["blade"]["structure"][
            "spar_cap_ps"
        ]["n_opt"]
        self.n_opt_te_ss = n_opt_te_ss = opt_options["design_variables"]["blade"]["structure"]["te_ss"]["n_opt"]
        self.n_opt_te_ps = n_opt_te_ps = opt_options["design_variables"]["blade"]["structure"]["te_ps"]["n_opt"]

        # Inputs
        self.add_input(
            "s",
            val=np.zeros(n_span),
            desc="1D array of the non-dimensional spanwise grid defined along blade axis (0-blade root, 1-blade tip)",
        )
        self.add_input(
            "layer_thickness_original",
            val=np.zeros((n_layers, n_span)),
            units="m",
            desc="2D array of the thickness of the layers of the blade structure. The first dimension represents each layer, the second dimension represents each entry along blade span.",
        )
        # Blade spar suction side
        self.add_input(
            "s_opt_spar_cap_ss",
            val=np.zeros(n_opt_spar_cap_ss),
            desc="1D array of the non-dimensional spanwise grid defined along blade axis to optimize the blade spar cap suction side",
        )
        self.add_input(
            "spar_cap_ss_opt",
            val=np.ones(n_opt_spar_cap_ss),
            units="m",
            desc="1D array of the the blade spanwise distribution of the spar caps suction side being optimized",
        )
        # Blade spar suction side
        self.add_input(
            "s_opt_spar_cap_ps",
            val=np.zeros(n_opt_spar_cap_ps),
            desc="1D array of the non-dimensional spanwise grid defined along blade axis to optimize the blade spar cap pressure side",
        )
        self.add_input(
            "spar_cap_ps_opt",
            val=np.ones(n_opt_spar_cap_ps),
            units="m",
            desc="1D array of the the blade spanwise distribution of the spar caps pressure side being optimized",
        )
        # Blade TE suction side
        self.add_input(
            "s_opt_te_ss",
            val=np.zeros(n_opt_te_ss),
            desc="1D array of the non-dimensional spanwise grid defined along blade axis to optimize the blade trailing edge suction side",
        )
        self.add_input(
            "te_ss_opt",
            val=np.ones(n_opt_te_ss),
            units="m",
            desc="1D array of the the blade spanwise distribution of the trailing edges suction side being optimized",
        )
        # Blade TE suction side
        self.add_input(
            "s_opt_te_ps",
            val=np.zeros(n_opt_te_ps),
            desc="1D array of the non-dimensional spanwise grid defined along blade axis to optimize the blade trailing edge pressure side",
        )
        self.add_input(
            "te_ps_opt",
            val=np.ones(n_opt_te_ps),
            units="m",
            desc="1D array of the the blade spanwise distribution of the trailing edges pressure side being optimized",
        )

        # Outputs
        self.add_output(
            "layer_thickness_param",
            val=np.zeros((n_layers, n_span)),
            units="m",
            desc="2D array of the thickness of the layers of the blade structure after the parametrization. The first dimension represents each layer, the second dimension represents each entry along blade span.",
        )

    def compute(self, inputs, outputs):

        layer_name = self.options["rotorse_options"]["layer_name"]

        spar_cap_ss_name = self.options["rotorse_options"]["spar_cap_ss"]
        spar_cap_ps_name = self.options["rotorse_options"]["spar_cap_ps"]

        ss_before_ps = False
        opt_ss = self.opt_options["design_variables"]["blade"]["structure"]["spar_cap_ss"]["flag"]
        opt_ps = self.opt_options["design_variables"]["blade"]["structure"]["spar_cap_ss"]["flag"]
        for i in range(self.n_layers):
            if layer_name[i] == spar_cap_ss_name and opt_ss and opt_ps:
                opt_m_interp = np.interp(inputs["s"], inputs["s_opt_spar_cap_ss"], inputs["spar_cap_ss_opt"])
                ss_before_ps = True
            elif layer_name[i] == spar_cap_ps_name and opt_ss and opt_ps:
                if (
                    self.opt_options["design_variables"]["blade"]["structure"]["spar_cap_ps"]["equal_to_suction"]
                    == False
                ) or ss_before_ps == False:
                    opt_m_interp = np.interp(inputs["s"], inputs["s_opt_spar_cap_ps"], inputs["spar_cap_ps_opt"])
                else:
                    opt_m_interp = np.interp(inputs["s"], inputs["s_opt_spar_cap_ss"], inputs["spar_cap_ss_opt"])
            else:
                opt_m_interp = inputs["layer_thickness_original"][i, :]

            outputs["layer_thickness_param"][i, :] = opt_m_interp

        te_ss_name = self.options["rotorse_options"]["te_ss"]
        te_ps_name = self.options["rotorse_options"]["te_ps"]

        ss_before_ps = False
        opt_ss = self.opt_options["design_variables"]["blade"]["structure"]["te_ss"]["flag"]
        opt_ps = self.opt_options["design_variables"]["blade"]["structure"]["te_ss"]["flag"]
        for i in range(self.n_layers):
            if layer_name[i] == te_ss_name and opt_ss and opt_ps:
                opt_m_interp = np.interp(inputs["s"], inputs["s_opt_te_ss"], inputs["te_ss_opt"])
                ss_before_ps = True
            elif layer_name[i] == te_ps_name and opt_ss and opt_ps:
                if (
                    self.opt_options["design_variables"]["blade"]["structure"]["te_ps"]["equal_to_suction"] == False
                ) or ss_before_ps == False:
                    opt_m_interp = np.interp(inputs["s"], inputs["s_opt_te_ps"], inputs["te_ps_opt"])
                else:
                    opt_m_interp = np.interp(inputs["s"], inputs["s_opt_te_ss"], inputs["te_ss_opt"])
            else:
                opt_m_interp = outputs["layer_thickness_param"][i, :]

            outputs["layer_thickness_param"][i, :] = opt_m_interp


class ComputeReynolds(om.ExplicitComponent):
    def initialize(self):
        self.options.declare("n_span")

    def setup(self):
        n_span = self.options["n_span"]

        self.add_input("rho", val=0.0, units="kg/m**3")
        self.add_input("mu", val=1.81e-5, units="kg/(m*s)", desc="Dynamic viscosity of air")
        self.add_input("local_airfoil_velocities", val=np.zeros((n_span)), units="m/s")
        self.add_input("chord", val=np.zeros((n_span)), units="m")
        self.add_output("Re", val=np.zeros((n_span)), ref=1.0e6)

    def compute(self, inputs, outputs):
        outputs["Re"] = np.nan_to_num(
            inputs["rho"] * inputs["local_airfoil_velocities"] * inputs["chord"] / inputs["mu"]
        )
